import re

from django import template


register = template.Library()


class_matcher = re.compile('(class="[^"]+)(")')
tag_matcher = re.compile(r"<([a-z0-9]+)([ >])")


def mark_current(navigation, current):
    """
    searches for links and marks the links 'active' (add the class "active")
    depending on the request.path.

    usage::

        {% mark_current request.path %}
          <li><a href="/">Home</a></li>
          <li><a href="/products/">Products</a></li>
        {% endmark_current %}

    resulting in (assuming request.path == '/products/')::

        <li><a href="/">Home</a></li>
        <li class="active"><a class="active" href="/products/">Products</a>\
            </li>

    .. note::

        This tag works line by line. so if <li> and <a ..> are on the same
        line, both get marked.
    """
    current_tokens = current.split("/")
    tokens = []

    low = 1
    if len(current_tokens) > 2:
        low = 2

    for i in range(low, len(current_tokens)):
        tokens.append(re.escape("/".join(current_tokens[0:i]) + "/"))

    current_matcher = re.compile('href="(' + "|".join(tokens) + ')"')

    lines = navigation.splitlines()
    out = []
    for line in lines:
        if current_matcher.search(line):
            if class_matcher.search(line):
                out.append(class_matcher.sub(r"\1 active\2", line))
            else:
                out.append(tag_matcher.sub(r'<\1 class="active"\2', line))
        else:
            out.append(line)

    return "\n".join(out)


def do_mark_current(parser, token):
    try:
        tag_name, url_variable = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires one argument" % token.contents.split()[0]
        )

    nodelist = parser.parse(("endmark_current",))
    parser.delete_first_token()
    return MarkCurrentNode(nodelist, url_variable)


class MarkCurrentNode(template.Node):
    def __init__(self, nodelist, url_variable):
        self.nodelist = nodelist
        self.url_variable = template.Variable(url_variable)

    def render(self, context):
        output = self.nodelist.render(context)
        return mark_current(output, self.url_variable.resolve(context))


register.tag("mark_current", do_mark_current)



# Generated by Django 2.2rc1 on 2019-03-30 12:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("projects", "0007_auto_20190321_1449")]

    operations = [
        migrations.AddField(
            model_name="service",
            name="is_logging_prohibited",
            field=models.BooleanField(
                default=False,
                help_text="Use this for service entries which are only used for budgeting.",
                verbose_name="is logging prohibited",
            ),
        )
    ]



from django.urls import re_path

from workbench import generic
from workbench.awt.forms import AbsenceForm, AbsenceSearchForm
from workbench.awt.models import Absence


urlpatterns = [
    re_path(
        r"^$",
        generic.ListView.as_view(
            model=Absence,
            search_form_class=AbsenceSearchForm,
            # show_create_button=False,
        ),
        name="awt_absence_list",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=Absence),
        name="awt_absence_detail",
    ),
    re_path(
        r"^create/$",
        generic.CreateView.as_view(model=Absence, form_class=AbsenceForm),
        name="awt_absence_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(model=Absence, form_class=AbsenceForm),
        name="awt_absence_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=Absence, template_name="modal_confirm_delete.html"
        ),
        name="awt_absence_delete",
    ),
]



# Generated by Django 2.2.3 on 2019-07-15 15:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contacts", "0008_auto_20190703_0801")]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="is_private_person",
            field=models.BooleanField(default=False, verbose_name="is private person"),
        )
    ]



import datetime as dt
import io

from django.core.management import BaseCommand

from workbench.credit_control.reporting import paid_debtors_zip


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--year",
            type=int,
            default=dt.date.today().year,
            help="The year for which to export debtors (defaults to %(default)s)",
        )
        parser.add_argument("target", type=str)

    def handle(self, **options):
        date_range = [dt.date(options["year"], 1, 1), dt.date(options["year"], 12, 31)]
        with io.open(options["target"], "wb") as f:
            paid_debtors_zip(date_range, file=f)



# Generated by Django 3.0 on 2019-12-07 12:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0004_auto_20190327_0945"),
    ]

    operations = [
        migrations.CreateModel(
            name="WorkingTimeModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
            ],
            options={
                "verbose_name": "working time model",
                "verbose_name_plural": "working time models",
                "ordering": ["name"],
            },
        ),
        migrations.AddField(
            model_name="year",
            name="working_time_model",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="awt.WorkingTimeModel",
                verbose_name="working time model",
            ),
        ),
    ]



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST

from workbench.deals.models import Deal
from workbench.offers.forms import OfferAutocompleteForm


def add_offer(request, pk):
    deal = get_object_or_404(Deal, pk=pk)
    form = OfferAutocompleteForm(request.POST if request.method == "POST" else None)
    if form.is_valid():
        offer = form.cleaned_data["offer"]
        deal.related_offers.add(offer)
        messages.success(
            request,
            _("Successfully added the offer %(offer)s as a related offer.")
            % {"offer": offer},
        )
        return HttpResponse("OK", status=201)

    return render(
        request,
        "generic/select_object.html",
        {"form": form, "title": _("Link offer to %(deal)s") % {"deal": deal}},
    )


@require_POST
def remove_offer(request, pk):
    deal = get_object_or_404(Deal, pk=pk)
    form = OfferAutocompleteForm(request.POST)
    if form.is_valid():
        offer = form.cleaned_data["offer"]
        deal.related_offers.remove(offer)
        messages.success(
            request,
            _("Successfully removed the offer %(offer)s as a related offer.")
            % {"offer": offer},
        )
    return redirect(deal)



# Generated by Django 2.1.7 on 2019-03-04 21:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("logbook", "0001_initial"),
        ("projects", "0001_initial"),
        ("invoices", "0002_invoice_project"),
    ]

    operations = [
        migrations.AddField(
            model_name="loggedhours",
            name="service",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="loggedhours",
                to="projects.Service",
                verbose_name="service",
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="created_by",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
                verbose_name="created by",
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="invoice",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="invoices.Invoice",
                verbose_name="invoice",
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="project",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="loggedcosts",
                to="projects.Project",
                verbose_name="project",
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="service",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="loggedcosts",
                to="projects.Service",
                verbose_name="service",
            ),
        ),
        migrations.AddIndex(
            model_name="loggedhours",
            index=models.Index(
                fields=["-rendered_on"], name="logbook_log_rendere_ea492e_idx"
            ),
        ),
    ]



import datetime as dt

from django import forms
from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.contacts.forms import PostalAddressSelectionForm
from workbench.contacts.models import Organization, Person
from workbench.invoices.models import Invoice, RecurringInvoice, Service
from workbench.logbook.models import LoggedCost, LoggedHours
from workbench.services.models import ServiceType
from workbench.tools.formats import Z2, currency, hours, local_date_format
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea
from workbench.tools.pdf import pdf_response
from workbench.tools.validation import in_days
from workbench.tools.xlsx import WorkbenchXLSXDocument


class InvoiceSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=(
            ("", _("All states")),
            ("open", _("Open")),
            (_("Exact"), Invoice.STATUS_CHOICES),
        ),
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    org = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("s") == "open":
            queryset = queryset.open()
        elif data.get("s"):
            queryset = queryset.filter(status=data.get("s"))
        queryset = self.apply_renamed(queryset, "org", "customer")
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related(
            "customer", "contact__organization", "owned_by", "project__owned_by"
        )

    def response(self, request, queryset):
        if request.GET.get("export") == "pdf":
            if not queryset.exists():
                messages.warning(request, _("No invoices found."))
                return HttpResponseRedirect("?error=1")

            count = queryset.count()
            if count > settings.BATCH_MAX_ITEMS:
                messages.error(
                    request, _("%s invoices in selection, that's too many.") % count
                )
                return HttpResponseRedirect("?error=1")

            pdf, response = pdf_response(
                "invoices",
                as_attachment=request.GET.get("disposition") == "attachment",
            )
            for invoice in queryset:
                pdf.init_letter()
                pdf.process_invoice(invoice)
                pdf.restart()
            pdf.generate()
            return response

        if request.GET.get("export") == "xlsx":
            xlsx = WorkbenchXLSXDocument()
            xlsx.table_from_queryset(queryset)
            return xlsx.to_response("invoices.xlsx")


class InvoiceForm(PostalAddressSelectionForm):
    user_fields = default_to_current_user = ("owned_by",)

    class Meta:
        model = Invoice
        fields = (
            "contact",
            "customer",
            "invoiced_on",
            "due_on",
            "title",
            "description",
            "service_period_from",
            "service_period_until",
            "owned_by",
            "status",
            "closed_on",
            "payment_notice",
            "postal_address",
            "type",
            "subtotal",
            "third_party_costs",
            "discount",
            "liable_to_vat",
            "show_service_details",
        )
        widgets = {
            "customer": Autocomplete(model=Organization),
            "contact": Autocomplete(model=Person, params={"only_employees": "on"}),
            "status": forms.RadioSelect,
            "description": Textarea,
            "payment_notice": Textarea({"rows": 2}),
            "postal_address": Textarea,
            "type": forms.RadioSelect,
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["type"].choices = Invoice.TYPE_CHOICES
        self.fields["type"].disabled = True

        if self.instance.project:
            self.fields["customer"].disabled = True
            self.fields["customer"].help_text = _("Determined by project.")

        if self.instance.type == self.instance.DOWN_PAYMENT:
            self.fields["subtotal"].label = _("Down payment")
            self.fields.pop("service_period_from")
            self.fields.pop("service_period_until")

        if self.instance.type == self.instance.SERVICES:
            self.fields["subtotal"].disabled = True
            self.fields["third_party_costs"].disabled = True
            self.fields["subtotal"].help_text = _("Calculated from invoice services.")
            self.fields["third_party_costs"].help_text = _(
                "Calculated from invoice services."
            )

        else:
            self.fields.pop("show_service_details")

        if self.instance.type != Invoice.DOWN_PAYMENT and self.instance.project_id:
            eligible_down_payment_invoices = Invoice.objects.invoiced().filter(
                Q(project=self.instance.project),
                Q(type=Invoice.DOWN_PAYMENT),
                Q(down_payment_applied_to__isnull=True)
                | Q(down_payment_applied_to=self.instance),
            )
            if eligible_down_payment_invoices:
                self.fields["apply_down_payment"] = forms.ModelMultipleChoiceField(
                    label=_("Apply down payment invoices"),
                    queryset=eligible_down_payment_invoices,
                    widget=forms.CheckboxSelectMultiple,
                    required=False,
                    initial=Invoice.objects.filter(
                        down_payment_applied_to=self.instance
                    ).values_list("id", flat=True)
                    if self.instance.pk
                    else [],
                )
                self.fields["apply_down_payment"].choices = [
                    (
                        invoice.id,
                        format_html(
                            "{}<br/>{}, {}",
                            invoice.__html__(),
                            invoice.pretty_total_excl,
                            invoice.pretty_status,
                        ),
                    )
                    for invoice in eligible_down_payment_invoices
                ]

        self.order_fields(
            field
            for field in list(self.fields)
            if field
            not in {
                "subtotal",
                "discount",
                "apply_down_payment",
                "liable_to_vat",
                "show_service_details",
            }
        )

        self.add_postal_address_selection_if_empty(
            person=self.instance.contact, for_billing=True
        )

    def _is_status_unexpected(self, to_status):
        from_status = self.instance._orig_status

        if from_status == to_status:
            return False
        if from_status > to_status or from_status >= Invoice.PAID:
            return True
        return False

    def clean(self):
        data = super().clean()
        s_dict = dict(Invoice.STATUS_CHOICES)

        if self.instance.status > self.instance.IN_PREPARATION:
            if set(self.changed_data) - {
                "status",
                "closed_on",
                "payment_notice",
                "third_party_costs",
            }:
                self.add_warning(
                    _(
                        "You are attempting to change %(fields)s."
                        " I am trying to prevent unintentional changes."
                        " Are you sure?"
                    )
                    % {
                        "fields": ", ".join(
                            "'%s'" % self.fields[field].label
                            for field in self.changed_data
                        )
                    },
                    code="maybe-unintentional-invoice-change",
                )

        if self._is_status_unexpected(data["status"]):
            self.add_warning(
                _("Moving status from '%(from)s' to '%(to)s'. Are you sure?")
                % {
                    "from": s_dict[self.instance._orig_status],
                    "to": s_dict[data["status"]],
                },
                code="status-unexpected",
            )

        if (
            self.instance._orig_status == Invoice.IN_PREPARATION
            and data["status"] > Invoice.IN_PREPARATION
            and data.get("invoiced_on")
            and data["invoiced_on"] < in_days(-3)
        ):
            self.add_warning(
                _(
                    "Moving status from '%(from)s' to '%(to)s' with an"
                    " invoice date in the past (%(date)s)."
                    " This may be correct but maybe it's not what you want."
                )
                % {
                    "from": s_dict[Invoice.IN_PREPARATION],
                    "to": s_dict[data["status"]],
                    "date": local_date_format(data["invoiced_on"]),
                },
                code="invoiced-in-past",
            )

        if data["status"] >= Invoice.PAID and not data["closed_on"]:
            data["closed_on"] = dt.date.today()

        if self.instance.closed_on and data["status"] < Invoice.PAID:
            if self.should_ignore_warnings():
                data["closed_on"] = None
            else:
                self.add_warning(
                    _(
                        "You are attempting to set status to '%(to)s',"
                        " but the invoice has already been closed on %(closed)s."
                        " Are you sure?"
                    )
                    % {
                        "to": s_dict[data["status"]],
                        "closed": local_date_format(self.instance.closed_on),
                    },
                    code="status-change-but-already-closed",
                )

        return data

    def save(self):
        instance = super().save(commit=False)

        if instance.type in (instance.FIXED, instance.DOWN_PAYMENT):
            instance.subtotal = self.cleaned_data["subtotal"]
            instance.down_payment_total = Z2

        if "apply_down_payment" in self.cleaned_data:
            if not instance.pk:
                instance.save()

            instance.down_payment_invoices.set(
                self.cleaned_data.get("apply_down_payment")
            )
            instance.down_payment_total = sum(
                (
                    invoice.total_excl_tax
                    for invoice in self.cleaned_data.get("apply_down_payment")
                ),
                Z2,
            )

        instance.save()
        return instance


class CreateProjectInvoiceForm(InvoiceForm):
    def __init__(self, *args, **kwargs):
        self.project = kwargs.pop("project")
        type = kwargs["request"].GET.get("type")
        kwargs["instance"] = Invoice(
            customer=self.project.customer,
            contact=self.project.contact,
            project=self.project,
            title=self.project.title,
            description=self.project.description,
            type=type,
            third_party_costs=Z2 if type == Invoice.SERVICES else None,
        )

        super().__init__(*args, **kwargs)

        # Hide those fields when creating invoices
        self.fields["status"].disabled = True
        self.fields["closed_on"].disabled = True

        invoice_type = self.request.GET.get("type")
        self.fields["type"].initial = self.instance.type = invoice_type

        if invoice_type == "services":
            self.add_services_field()
            self.fields.pop("service_period_from")
            self.fields.pop("service_period_until")

        if (
            self.request.method == "GET"
            and self.project.invoices.filter(status=Invoice.IN_PREPARATION).exists()
        ):
            messages.warning(
                self.request, _("This project already has an invoice in preparation.")
            )

    def add_services_field(self):
        source = self.request.GET.get("source")
        if source == "logbook":

            def amount(row):
                if row["service"].effort_rate is not None:
                    return currency(row["not_archived"])
                elif row["logged_hours"]:
                    return format_html(
                        '{} <small class="bg-warning px-1">{}</small>',
                        currency(row["not_archived"]),
                        _("%s logged but no hourly rate defined.")
                        % hours(row["logged_hours"]),
                    )
                return currency(row["logged_cost"])

        else:

            def amount(row):
                return currency(row["service"].service_cost)

        choices = []
        for offer, offer_data in self.project.grouped_services["offers"]:
            choices.append(
                (
                    format_html(
                        "<u>{}</u><br>"
                        '<div class="form-check">'
                        '<input type="checkbox" data-toggle-following>'
                        "{}"
                        "</div>",
                        offer if offer else _("Not offered yet"),
                        _("Select all"),
                    ),
                    [
                        (
                            row["service"].id,
                            format_html(
                                '<div class="mb-2"><strong>{}</strong>'
                                "<br>{}{}</div>",
                                row["service"].title,
                                format_html("{}<br>", row["service"].description)
                                if row["service"].description
                                else "",
                                amount(row),
                            ),
                        )
                        for row in offer_data["services"]
                    ],
                )
            )
        self.fields["selected_services"] = forms.MultipleChoiceField(
            choices=choices,
            label=_("services"),
            widget=forms.CheckboxSelectMultiple(attrs={"size": 30}),
        )

        self.fields["disable_logging"] = forms.TypedChoiceField(
            label=_("Disable logging on selected services from now on?"),
            choices=[
                (1, _("Yes, disable logging from now on")),
                (0, _("No, do not change anything")),
            ],
            coerce=lambda x: int(x),
            widget=forms.RadioSelect,
        )

    def save(self):
        if self.request.GET.get("type") != "services":
            return super().save()

        instance = super().save()
        services = self.project.services.filter(
            id__in=self.cleaned_data["selected_services"]
        )
        if self.request.GET.get("source") == "logbook":
            instance.create_services_from_logbook(services)
        else:
            instance.create_services_from_offer(services)
        if self.cleaned_data["disable_logging"]:
            services.update(allow_logging=False)
        return instance


class CreatePersonInvoiceForm(PostalAddressSelectionForm):
    user_fields = default_to_current_user = ("owned_by",)
    type = forms.ChoiceField(
        label=_("type"),
        choices=Invoice.TYPE_CHOICES,
        initial=Invoice.FIXED,
        disabled=True,
        widget=forms.RadioSelect,
        help_text=_("Other invoice types must be created directly on a project."),
    )

    class Meta:
        model = Invoice
        fields = (
            "contact",
            "customer",
            "invoiced_on",
            "due_on",
            "title",
            "description",
            "owned_by",
            "postal_address",
            "subtotal",
            "third_party_costs",
            "discount",
            "liable_to_vat",
        )
        widgets = {
            "customer": Autocomplete(model=Organization),
            "contact": Autocomplete(model=Person, params={"only_employees": "on"}),
            "description": Textarea,
            "postal_address": Textarea,
        }

    def __init__(self, *args, **kwargs):
        request = kwargs["request"]
        initial = kwargs.setdefault("initial", {})
        initial.update(
            {"subtotal": None, "third_party_costs": None}  # Invalid -- force input.
        )

        contact = None
        customer = None
        self.pre_form = False

        if pk := request.GET.get("contact"):
            try:
                contact = Person.objects.get(pk=pk)
            except (Person.DoesNotExist, TypeError, ValueError):
                self.pre_form = True
            else:
                initial.update({"customer": contact.organization, "contact": contact})

        elif pk := request.GET.get("customer"):
            try:
                customer = Organization.objects.get(pk=pk)
            except (Organization.DoesNotExist, TypeError, ValueError):
                self.pre_form = True
            else:
                initial.update({"customer": customer})

        else:
            self.pre_form = True

        super().__init__(*args, **kwargs)

        self.instance.type = Invoice.FIXED

        if self.pre_form:
            for field in list(self.fields):
                if field not in {"customer", "contact"}:
                    self.fields.pop(field)
        else:
            self.add_postal_address_selection_if_empty(
                person=contact, organization=customer, for_billing=True
            )


class InvoiceDeleteForm(ModelForm):
    class Meta:
        model = Invoice
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert (
            self.instance.status <= self.instance.IN_PREPARATION
        ), "Trying to delete an invoice not in preparation"
        if (
            LoggedHours.objects.filter(invoice_service__invoice=self.instance).exists()
            or LoggedCost.objects.filter(
                invoice_service__invoice=self.instance
            ).exists()
        ):
            self.add_warning(
                _(
                    "Logged services are linked with this invoice."
                    " They will be released when deleting this invoice."
                ),
                code="release-logged-services",
            )

        if self.instance.down_payment_invoices.exists():
            self.add_warning(
                _(
                    "Down payment invoices are subtracted from this invoice."
                    " Those down payments will be released when deleting this"
                    " invoice."
                ),
                code="release-down-payments",
            )

    def delete(self):
        LoggedHours.objects.filter(invoice_service__invoice=self.instance).update(
            invoice_service=None, archived_at=None
        )
        LoggedCost.objects.filter(invoice_service__invoice=self.instance).update(
            invoice_service=None, archived_at=None
        )
        self.instance.down_payment_invoices.update(down_payment_applied_to=None)
        self.instance.delete()


class ServiceForm(ModelForm):
    service_type = forms.ModelChoiceField(
        ServiceType.objects.all(),
        label=ServiceType._meta.verbose_name,
        required=False,
        help_text=_("Optional, but useful for quickly filling the fields below."),
    )

    class Meta:
        model = Service
        fields = [
            "title",
            "description",
            "effort_type",
            "effort_hours",
            "effort_rate",
            "cost",
            "third_party_costs",
        ]
        widgets = {"description": Textarea}

    def __init__(self, *args, **kwargs):
        self.invoice = kwargs.pop("invoice", None) or kwargs["instance"].invoice
        super().__init__(*args, **kwargs)
        self.instance.invoice = self.invoice


class RecurringInvoiceSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=(("all", _("All states")), ("", _("Open")), ("closed", _("Closed"))),
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    org = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("s") == "":
            queryset = queryset.filter(
                Q(ends_on__isnull=True) | Q(ends_on__gte=dt.date.today())
            )
        elif data.get("s") == "closed":
            queryset = queryset.filter(
                Q(ends_on__isnull=False) & Q(ends_on__lt=dt.date.today())
            )
        queryset = self.apply_renamed(queryset, "org", "customer")
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related("customer", "contact__organization", "owned_by")


class RecurringInvoiceForm(PostalAddressSelectionForm):
    user_fields = default_to_current_user = ("owned_by",)

    class Meta:
        model = RecurringInvoice
        fields = (
            "contact",
            "customer",
            "title",
            "description",
            "owned_by",
            "postal_address",
            "starts_on",
            "ends_on",
            "periodicity",
            "create_invoice_on_day",
            "next_period_starts_on",
            "subtotal",
            "third_party_costs",
            "discount",
            "liable_to_vat",
            "create_project",
        )
        widgets = {
            "customer": Autocomplete(model=Organization),
            "contact": Autocomplete(model=Person, params={"only_employees": "on"}),
            "description": Textarea,
            "postal_address": Textarea,
            "periodicity": forms.RadioSelect,
        }

    def __init__(self, *args, **kwargs):
        instance = kwargs.get("instance")
        contact = None
        customer = None
        self.pre_form = False
        request = kwargs["request"]
        initial = kwargs.setdefault("initial", {})

        if pk := request.GET.get("copy"):
            try:
                invoice = RecurringInvoice.objects.get(pk=pk)
            except (RecurringInvoice.DoesNotExist, TypeError, ValueError):
                pass
            else:
                initial.update(
                    {
                        "customer": invoice.customer_id,
                        "contact": invoice.contact_id,
                        "title": invoice.title,
                        "description": invoice.description,
                        "owned_by": invoice.owned_by_id,
                        "postal_address": invoice.postal_address,
                        "periodicity": invoice.periodicity,
                        "create_invoice_on_day": invoice.create_invoice_on_day,
                        "subtotal": invoice.subtotal,
                        "discount": invoice.discount,
                        "third_party_costs": invoice.third_party_costs,
                        "liable_to_vat": invoice.liable_to_vat,
                    }
                )

        elif pk := request.GET.get("contact"):
            try:
                contact = Person.objects.get(pk=pk)
            except (Person.DoesNotExist, TypeError, ValueError):
                self.pre_form = True
            else:
                initial.update({"customer": contact.organization, "contact": contact})

        elif pk := request.GET.get("customer"):
            try:
                customer = Organization.objects.get(pk=pk)
            except (Organization.DoesNotExist, TypeError, ValueError):
                self.pre_form = True
            else:
                initial.update({"customer": customer})

        elif not instance:
            self.pre_form = True

        if not self.pre_form and not instance:
            kwargs["instance"] = RecurringInvoice(subtotal=None, third_party_costs=None)

        super().__init__(*args, **kwargs)

        if self.pre_form:
            for field in list(self.fields):
                if field not in {"customer", "contact"}:
                    self.fields.pop(field)

        else:
            self.fields["periodicity"].choices = RecurringInvoice.PERIODICITY_CHOICES
            self.fields["next_period_starts_on"].disabled = True

            self.add_postal_address_selection_if_empty(
                person=self.instance.contact or contact,
                organization=customer,
                for_billing=True,
            )



import re

from django import forms
from django.utils.html import format_html, mark_safe
from django.utils.translation import gettext, gettext_lazy as _

from workbench.credit_control.models import CreditEntry, Ledger
from workbench.invoices.models import Invoice
from workbench.tools.formats import currency, local_date_format
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea


class CreditEntrySearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=(
            ("", _("All states")),
            ("pending", _("Pending")),
            ("processed", _("Processed")),
        ),
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    ledger = forms.ModelChoiceField(
        Ledger.objects.all(),
        required=False,
        empty_label=_("All ledgers"),
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        if data.get("s") == "pending":
            queryset = queryset.pending()
        elif data.get("s") == "processed":
            queryset = queryset.processed()
        queryset = self.apply_simple(queryset, "ledger")
        return queryset.select_related(
            "invoice__project", "invoice__owned_by", "ledger"
        )


class CreditEntryForm(ModelForm):
    class Meta:
        model = CreditEntry
        fields = [
            "ledger",
            "reference_number",
            "value_date",
            "total",
            "payment_notice",
            "invoice",
            "notes",
        ]
        widgets = {"invoice": Autocomplete(model=Invoice), "notes": Textarea}

    def save(self):
        instance = super().save()
        if instance.invoice and instance.invoice.status != instance.invoice.PAID:
            instance.invoice.status = instance.invoice.PAID
            instance.invoice.closed_on = instance.value_date
            instance.invoice.payment_notice = instance.payment_notice
            instance.invoice.save()
        return instance


class AccountStatementUploadForm(Form):
    ledger = CreditEntry._meta.get_field("ledger").formfield(widget=forms.RadioSelect)
    statement = forms.FileField(label=_("Account statement"))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["ledger"].choices = [
            (ledger.id, str(ledger)) for ledger in Ledger.objects.all()
        ]

    def clean(self):
        data = super().clean()
        if data.get("statement") and data.get("ledger"):
            try:
                self.statement_list = data["ledger"].parse_fn(data["statement"].read())
            except Exception as exc:
                raise forms.ValidationError(
                    _(
                        "Error while parsing the statement."
                        " Did you upload a valid CSV file? (Technical error: %s)"
                    )
                    % exc
                )

            reference_numbers = [
                entry["reference_number"] for entry in self.statement_list
            ]

            if (
                reference_numbers
                and not CreditEntry.objects.filter(
                    ledger=data["ledger"], reference_number__in=reference_numbers
                ).exists()
            ):
                self.add_warning(
                    _(
                        "The uploaded list only contains new payments."
                        " This is somewhat surprising if you exported"
                        " the list from the correct bank account.",
                    ),
                    code="no-known-payments",
                )

        return data

    def save(self):
        created_entries = []
        ledger = self.cleaned_data["ledger"]

        for data in reversed(self.statement_list):  # From past to present
            reference_number = data.pop("reference_number")

            c, created = ledger.transactions.get_or_create(
                reference_number=reference_number, defaults=data
            )
            if created:
                created_entries.append(c)
        return created_entries


class AssignCreditEntriesForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")

        super().__init__(*args, **kwargs)

        self.entries = []
        for entry in CreditEntry.objects.reverse().filter(
            invoice__isnull=True, notes=""
        )[:20]:
            self.fields["entry_{}_invoice".format(entry.pk)] = forms.TypedChoiceField(
                label=format_html(
                    '<a href="{}" target="_blank"'
                    ' rel="noopener noreferrer">{}, {}: {}</a>',
                    entry.get_absolute_url(),
                    entry.total,
                    local_date_format(entry.value_date),
                    entry.payment_notice,
                ),
                choices=[(None, "----------")]
                + [
                    (
                        invoice.id,
                        mark_safe(
                            " ".join(
                                (
                                    format_html(
                                        '<span title="{}">', invoice.description
                                    ),
                                    format_html(
                                        "<strong>{}</strong>"
                                        if re.search(
                                            r"\b" + invoice.code + r"\b",
                                            entry.payment_notice,
                                        )
                                        else "{}",
                                        invoice,
                                    ),
                                    invoice.status_badge,
                                    "<br>",
                                    format_html(
                                        "{}",
                                        invoice.contact.name_with_organization
                                        if invoice.contact
                                        else invoice.customer,
                                    ),
                                    "<br>",
                                    format_html(
                                        "{} {}",
                                        _("invoiced on"),
                                        local_date_format(invoice.invoiced_on),
                                    )
                                    if invoice.invoiced_on
                                    else gettext("NO DATE YET"),
                                    "<br>",
                                    currency(invoice.total),
                                    format_html(
                                        "<br><span style='color:darkred'>{}: {}</span>",
                                        _("third party costs"),
                                        currency(invoice.third_party_costs),
                                    )
                                    if invoice.third_party_costs
                                    else "",
                                    "</span>",
                                )
                            )
                        ),
                    )
                    for invoice in Invoice.objects.open()
                    .filter(total=entry.total)
                    .select_related(
                        "contact__organization", "customer", "owned_by", "project"
                    )[:100]
                ],
                coerce=int,
                required=False,
                widget=forms.RadioSelect,
            )

            self.fields["entry_{}_notes".format(entry.pk)] = forms.CharField(
                widget=Textarea({"rows": 1}), label=_("notes"), required=False
            )

            self.entries.append(
                (
                    entry,
                    "entry_{}_invoice".format(entry.pk),
                    "entry_{}_notes".format(entry.pk),
                )
            )

    def save(self):
        for entry, invoice_field, notes_field in self.entries:
            entry.invoice_id = self.cleaned_data.get(invoice_field) or None
            entry.notes = self.cleaned_data.get(notes_field, "")
            entry.save()

            if entry.invoice and entry.invoice.status != entry.invoice.PAID:
                entry.invoice.status = entry.invoice.PAID
                entry.invoice.closed_on = entry.value_date
                entry.invoice.payment_notice = entry.payment_notice
                entry.invoice.save()



# Generated by Django 2.1.7 on 2019-03-21 13:49

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("offers", "0003_auto_20190304_2247")]

    operations = [
        migrations.RunSQL(search.drop_old_shit("offers_offer")),
        migrations.RunSQL(search.create_structure("offers_offer")),
        migrations.RunSQL(
            search.fts("offers_offer", ["title", "description", "postal_address"])
        ),
    ]



import datetime as dt
import time
from collections import defaultdict

from django import forms
from django.contrib import messages
from django.db.models.functions import Coalesce
from django.shortcuts import redirect, render
from django.utils.translation import gettext as _

from workbench.accounts.features import FEATURES
from workbench.accounts.models import Team, User
from workbench.awt.models import Absence, Year
from workbench.awt.pdf import annual_working_time_pdf
from workbench.awt.reporting import active_users, annual_working_time
from workbench.tools.forms import Form
from workbench.tools.validation import filter_form, monday


def annual_working_time_view(request):
    this_year = dt.date.today().year
    try:
        year = int(request.GET.get("year", this_year))
    except Exception:
        return redirect(".")

    user = request.GET.get("user")
    users = None
    if user == "active" and request.user.features[FEATURES.CONTROLLING]:
        users = active_users(year)
    elif user and user != "active":
        users = User.objects.filter(id=user)
    if not users:
        users = [request.user]
    statistics = annual_working_time(year, users=users)
    for user in statistics["months"].users_without_wtm:
        messages.warning(
            request,
            _(
                "No annual working time defined for user %(user)s"
                " with working time model %(working_time_model)s."
            )
            % {"user": user, "working_time_model": user.working_time_model},
        )

    if request.GET.get("export") == "pdf":
        return annual_working_time_pdf(statistics)

    return render(
        request,
        "awt/year_detail.html",
        {
            "overall": statistics["overall"],
            "statistics": statistics["statistics"],
            "object": year,
            "year": year,
            "years": sorted(
                Year.objects.filter(year__lte=dt.date.today().year)
                .order_by()
                .values_list("year", flat=True)
                .distinct(),
                reverse=True,
            ),
            "view": {"meta": Year._meta},
            "is_last_year": year == this_year - 1,
            "this_year": this_year,
        },
    )


class UserFilterForm(Form):
    team = forms.ModelChoiceField(
        Team.objects.all(), empty_label=_("Everyone"), label="", required=False
    )

    def users(self):
        data = self.cleaned_data
        queryset = User.objects.active()
        if data.get("team"):
            queryset = queryset.filter(teams=data.get("team"))
        return queryset


@filter_form(UserFilterForm)
def absence_calendar(request, form):
    users = form.users()

    dates = {dt.date.today()}
    absences = defaultdict(list)
    cutoff = monday() - dt.timedelta(days=7)
    queryset = Absence.objects.annotate(
        _ends_on=Coalesce("ends_on", "starts_on")
    ).filter(
        starts_on__lte=cutoff + dt.timedelta(days=366),
        _ends_on__gte=cutoff,
        user__in=users,
    )

    for absence in queryset:
        absences[absence.user_id].append(absence)
        dates.add(max(absence.starts_on, cutoff))
        dates.add(absence._ends_on)

    absences = [
        {
            "name": user.get_full_name(),
            "id": user.id,
            "absences": [
                {
                    "id": absence.id,
                    "reason": absence.reason,
                    "reasonDisplay": absence.get_reason_display(),
                    "startsOn": time.mktime(max(absence.starts_on, cutoff).timetuple())
                    * 1000,
                    "endsOn": time.mktime(
                        (absence.ends_on or absence.starts_on).timetuple()
                    )
                    * 1000,
                    "days": absence.days,
                    "description": absence.description,
                }
                for absence in absences[user.id]
            ],
        }
        for user in users
    ]

    return render(
        request,
        "awt/absence_calendar.html",
        {
            "absences_data": {
                "absencesByPerson": absences,
                "reasonList": Absence.REASON_CHOICES,
                "timeBoundaries": {
                    "start": time.mktime(monday(min(dates)).timetuple()) * 1000,
                    "end": time.mktime(max(dates).timetuple()) * 1000,
                },
                "monday": time.mktime(monday().timetuple()),
            },
            "form": form,
        },
    )



import datetime as dt

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.test.utils import override_settings
from django.utils import timezone

from time_machine import travel

from workbench import factories
from workbench.logbook.models import Break
from workbench.timer.models import Timestamp
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code
from workbench.tools.validation import in_days


def c(day, time):
    return timezone.make_aware(dt.datetime.combine(day, time))


class BreaksTest(TestCase):
    def test_list(self):
        """Filter form smoke test"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        brk = Break.objects.create(
            user=factories.UserFactory.create(),
            starts_at=c(dt.date.today(), dt.time(12, 0)),
            ends_at=c(dt.date.today(), dt.time(13, 0)),
        )

        code = check_code(self, "/logbook/breaks/")
        code("")
        code("user=-1")
        code("user={}".format(user.pk))
        code("export=xlsx")

        response = self.client.get(brk.urls["detail"])
        self.assertContains(response, str(brk))

    def test_valid_break(self):
        """Model validation of a valid break raises no exceptions"""
        brk = Break(
            user=factories.UserFactory.create(),
            starts_at=c(dt.date.today(), dt.time(12, 0)),
            ends_at=c(dt.date.today(), dt.time(13, 0)),
        )
        brk.full_clean()
        self.assertEqual(brk.timedelta.total_seconds(), 3600)

    def test_break_validation(self):
        """Model validation of an invalid break fails"""
        with self.assertRaises(ValidationError) as cm:
            Break(
                user=factories.UserFactory.create(),
                starts_at=c(dt.date.today(), dt.time(12, 0)),
                ends_at=c(dt.date.today(), dt.time(11, 0)),
            ).full_clean()

        self.assertEqual(
            list(cm.exception),
            [("ends_at", ["Breaks should end later than they begin."])],
        )

        with self.assertRaises(ValidationError) as cm:
            Break(
                user=factories.UserFactory.create(),
                starts_at=c(dt.date(2020, 1, 1), dt.time(12, 0)),
                ends_at=c(dt.date.today(), dt.time(11, 0)),
            ).full_clean()

        self.assertEqual(
            list(cm.exception),
            [("ends_at", ["Breaks must start and end on the same day."])],
        )

    @travel("2020-01-02 12:00")
    def test_break_form_initial(self):
        """Fields are initialized as expected"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get(Break.urls["create"])
        self.assertContains(response, 'value="2020-01-02"')

        response = self.client.get(Break.urls["create"] + "?day=2020-01-01")
        self.assertContains(response, 'value="2020-01-01"')
        self.assertNotContains(response, 'value="2020-01-02"')

        response = self.client.get(Break.urls["create"] + "?starts_at=09:00:00")
        self.assertContains(response, 'value="09:00:00"')

        user.timestamp_set.create(
            created_at=timezone.now() - dt.timedelta(days=1), type=Timestamp.STOP
        )

        response = self.client.get(Break.urls["create"])
        self.assertContains(response, 'value="2020-01-02"')
        self.assertNotContains(response, 'value="2020-01-01"')

    def test_break_from_timestamps(self):
        """The breaks form URL contains expected values when using timestamps"""
        user = factories.UserFactory.create()
        now = timezone.localtime(timezone.now()).replace(
            hour=9, minute=0, second=0, microsecond=0
        )
        user.timestamp_set.create(type=Timestamp.STOP, created_at=now)
        user.timestamp_set.create(
            type=Timestamp.STOP, created_at=now - dt.timedelta(seconds=900)
        )

        self.client.force_login(user)

        response = self.client.get("/timestamps/")
        self.assertContains(response, "starts_at=08%3A45%3A00")
        self.assertContains(response, "ends_at=09%3A00%3A00")

    def test_day_validation(self):
        """Some day values are not allowed"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.post(
            Break.urls["create"],
            {
                "modal-day": "",
                "modal-starts_at": "12:00",
                "modal-ends_at": "12:45",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "This field is required.")

        response = self.client.post(
            Break.urls["create"],
            {
                "modal-day": dt.date.today().isoformat(),
                "modal-starts_at": "12:00",
                "modal-ends_at": "12",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "Enter a valid time.")

        response = self.client.post(
            Break.urls["create"],
            {
                "modal-day": "2010-01-01",
                "modal-starts_at": "12:00",
                "modal-ends_at": "12:45",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "Breaks have to be logged promptly.")

        response = self.client.post(
            Break.urls["create"],
            {
                "modal-day": "9999-01-01",
                "modal-starts_at": "12:00",
                "modal-ends_at": "12:45",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "That&#x27;s too far in the future.")

        response = self.client.post(
            Break.urls["create"],
            {
                "modal-day": dt.date.today().isoformat(),
                "modal-starts_at": "12:00",
                "modal-ends_at": "12:45",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

    def test_break_form_save_assigns_timestamp(self):
        """Creating a break from the timestamps button sets the logged_break attr"""
        user = factories.UserFactory.create()
        t = user.timestamp_set.create(type=Timestamp.STOP)

        self.client.force_login(user)

        response = self.client.post(
            Break.urls["create"] + "?timestamp={}".format(t.pk),
            {
                "modal-day": dt.date.today().isoformat(),
                "modal-starts_at": "12:00",
                "modal-ends_at": "12:45",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        brk = user.breaks.get()
        t.refresh_from_db()

        self.assertEqual(t.logged_break, brk)
        self.assertEqual(t.type, Timestamp.STOP)  # Type hasn't changed
        self.assertEqual(Timestamp.objects.count(), 1)

        t2 = user.timestamp_set.create(type=Timestamp.STOP)
        response = self.client.post(
            brk.urls["update"] + "?timestamp={}".format(t2.pk),
            {
                "modal-day": dt.date.today().isoformat(),
                "modal-starts_at": "12:00",
                "modal-ends_at": "12:45",
                "modal-description": "",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        t2.refresh_from_db()
        self.assertIsNone(t2.logged_break)  # No reassignment

    def test_update_delete_forbidden(self):
        """Updating and deleting of others' breaks is not allowed"""
        brk = Break.objects.create(
            user=factories.UserFactory.create(),
            starts_at=c(dt.date.today(), dt.time(12, 0)),
            ends_at=c(dt.date.today(), dt.time(13, 0)),
        )

        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get(
            brk.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "Cannot modify breaks of other users.")

        response = self.client.get(
            brk.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(response, "Cannot modify breaks of other users.")

    def test_break_warning(self):
        """Various places show a take-a-break warning if too much work and no break"""
        service = factories.ServiceFactory.create()
        user = service.project.owned_by
        self.client.force_login(user)

        factories.LoggedHoursFactory.create(rendered_by=user, hours=5)
        Break.objects.create(
            user=user,
            starts_at=c(dt.date.today(), dt.time(12, 0)),
            ends_at=c(dt.date.today(), dt.time(12, 5)),
        )

        data = {
            "modal-rendered_by": user.id,
            "modal-rendered_on": dt.date.today().isoformat(),
            "modal-service": service.id,
            "modal-hours": "2.0",
            "modal-description": "Test",
        }

        with override_settings(FEATURES={"skip_breaks": False}):
            response = self.client.post(
                service.project.urls["createhours"],
                data,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            )
            self.assertContains(response, "You should take")

            self.assertIsNotNone(user.take_a_break_warning(add=10))
            self.assertIsNotNone(user.take_a_break_warning(add=3))
            self.assertIsNotNone(user.take_a_break_warning(add=1))
            self.assertIsNone(user.take_a_break_warning(add=0))

            self.assertIsNone(user.take_a_break_warning(add=1, day=in_days(-1)))

            response = self.client.post(
                service.project.urls["createhours"],
                {**data, WarningsForm.ignore_warnings_id: "take-a-break"},
                HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            )
            self.assertEqual(response.status_code, 201)

        # With skip_breaks=True, everything just works
        response = self.client.post(
            service.project.urls["createhours"],
            {**data, "modal-hours": "3.0"},  # No duplicate
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        # Now the message also appears by default
        with override_settings(FEATURES={"skip_breaks": False}):
            response = self.client.get("/")
            self.assertContains(response, "You should take")



# Generated by Django 3.0.6 on 2020-05-11 11:24

import datetime as dt

from django.db import migrations, models
from django.utils import timezone


def forwards(apps, schema_editor):
    for brk in apps.get_model("logbook", "break").objects.all():
        brk._starts_at = timezone.make_aware(
            dt.datetime.combine(brk.day, brk.starts_at)
        )
        brk._ends_at = timezone.make_aware(dt.datetime.combine(brk.day, brk.ends_at))
        brk.save()


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0017_break"),
    ]

    operations = [
        migrations.AddField(
            model_name="break",
            name="_ends_at",
            field=models.DateTimeField(null=True, verbose_name="ends_at"),
        ),
        migrations.AddField(
            model_name="break",
            name="_starts_at",
            field=models.DateTimeField(null=True, verbose_name="starts_at"),
        ),
        migrations.RunPython(forwards),
    ]



from django import forms
from django.conf import settings
from django.utils.translation import gettext as _

from workbench.accounts.features import FEATURES
from workbench.accounts.models import Team, User
from workbench.awt.models import WorkingTimeModel
from workbench.tools.forms import Form, ModelForm


class UserSearchForm(Form):
    def filter(self, queryset):
        return queryset.filter(is_active=True)


class UserForm(ModelForm):
    class Meta:
        model = User
        fields = (
            "_full_name",
            "_short_name",
            "language",
            "email",
            "working_time_model",
            "planning_hours_per_day",
        )
        widgets = {
            "language": forms.RadioSelect,
            "working_time_model": forms.RadioSelect,
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["email"].disabled = True
        self.fields["email"].help_text = _(
            "Contact your administrator to change this value."
        )
        self.fields["language"].choices = settings.LANGUAGES
        self.fields["working_time_model"].choices = [
            (wtm.id, str(wtm)) for wtm in WorkingTimeModel.objects.all()
        ]

        if self.instance.pk:
            self.fields["working_time_model"].disabled = True
            self.fields["working_time_model"].help_text = _(
                "Contact your administrator to change this value."
            )

        user = self.request.user
        if not hasattr(user, "features") or not user.features[FEATURES.PLANNING]:
            self.fields.pop("planning_hours_per_day")


class TeamSearchForm(Form):
    def filter(self, queryset):
        return queryset


class TeamForm(ModelForm):
    class Meta:
        model = Team
        fields = ["name", "members"]
        widgets = {"members": forms.CheckboxSelectMultiple}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        candidates = User.objects.active()
        if self.instance.pk:
            candidates |= self.instance.members.all()
        self.fields["members"].queryset = candidates.distinct()



from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.contrib import admin
from django.shortcuts import redirect, render
from django.urls import include, path, re_path
from django.views.i18n import JavaScriptCatalog

from workbench import views


admin.site.enable_nav_sidebar = False


urlpatterns = [
    re_path(r"^$", views.start),
    re_path(r"^404/$", render, {"template_name": "404.html"}),
    re_path(
        r"^shortcuts/$", render, {"template_name": "shortcuts.html"}, name="shortcuts"
    ),
    re_path(r"^admin/", admin.site.urls),
    re_path(r"", include("workbench.accounts.urls")),
    re_path(r"^contacts/", include("workbench.contacts.urls")),
    re_path(r"^logbook/", include("workbench.logbook.urls")),
    re_path(r"^absences/", include("workbench.awt.urls")),
    re_path(r"^projects/", include("workbench.projects.urls")),
    re_path(r"^invoices/", include("workbench.invoices.urls")),
    re_path(r"^recurring-invoices/", include("workbench.invoices.recurring_urls")),
    re_path(r"^credit-control/", include("workbench.credit_control.urls")),
    re_path(r"^expenses/", include("workbench.expenses.urls")),
    re_path(r"^deals/", include("workbench.deals.urls")),
    re_path(r"^planning/", include("workbench.planning.urls")),
    re_path(r"^search/$", views.search, name="search"),
    re_path(r"^history/(\w+)/(\w+)/([0-9]+)/$", views.history, name="history"),
    re_path(r"^report/", include("workbench.reporting.urls")),
    re_path(r"", include("workbench.timer.urls")),
    re_path(r"^notes/", include("workbench.notes.urls")),
    # Legacy URL redirects
    re_path(
        r"^projects/projects/([0-9]+)/$",
        lambda request, pk: redirect("projects_project_detail", pk=pk),
    ),
    re_path(
        r"^offers/([0-9]+)/$",
        lambda request, pk: redirect("offers_offer_detail", pk=pk),
    ),
]

urlpatterns += i18n_patterns(
    path("jsi18n/", JavaScriptCatalog.as_view(), name="javascript-catalog"),
)

if settings.DEBUG_TOOLBAR:  # pragma: no cover
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns

    import debug_toolbar

    urlpatterns = (
        urlpatterns
        + [re_path(r"^__debug__/", include(debug_toolbar.urls))]
        + staticfiles_urlpatterns()
    )



from collections import defaultdict
from itertools import chain, groupby

from django.db import connections
from django.db.models import Sum

from workbench.accounts.models import User
from workbench.logbook.models import LoggedCost
from workbench.projects.models import Project
from workbench.tools.formats import Z1, Z2


LABOR_COSTS_SQL = """
select
    p.id,
    hourly_labor_costs,
    green_hours_target,
    lh.rendered_by_id,
    sum(lh.hours) as hours
from logbook_loggedhours lh
left join projects_service ps on lh.service_id=ps.id
left join projects_project p on ps.project_id=p.id
left outer join lateral (select * from awt_employment) as costs on
lh.rendered_by_id=costs.user_id
and lh.rendered_on >= costs.date_from
and lh.rendered_on <= costs.date_until
where %s
group by p.id, hourly_labor_costs, green_hours_target, lh.rendered_by_id
"""

REVENUE_SQL = """
with sq as (
    select
        p.id,
        greatest(
            coalesce(hourly_labor_costs, 0),
            coalesce(ps.effort_rate, 0)
        ) as min_effort_rate,
        sum(lh.hours) as hours
    from logbook_loggedhours lh
    left join projects_service ps on lh.service_id=ps.id
    left join projects_project p on ps.project_id=p.id
    left outer join lateral (select * from awt_employment) as costs on
    lh.rendered_by_id=costs.user_id
    and lh.rendered_on >= costs.date_from
    and lh.rendered_on <= costs.date_until
    where %s
    group by p.id, min_effort_rate
)
select id, sum(min_effort_rate * hours) from sq group by id
"""

USER_KEYS = [
    "hours",
    "hours_with_rate_undefined",
    "costs",
    "costs_with_green_hours_target",
]
PROJECT_KEYS = USER_KEYS + ["third_party_costs", "revenue"]


def _labor_costs_by_project_id(date_range, *, project=None, cost_center=None):
    projects = defaultdict(
        lambda: {
            "hours": Z1,
            "hours_with_rate_undefined": Z1,
            "costs": Z2,
            "costs_with_green_hours_target": Z2,
            "third_party_costs": Z2,
            "revenue": Z2,
            "by_user": defaultdict(
                lambda: {
                    "hours": Z1,
                    "hours_with_rate_undefined": Z1,
                    "costs": Z2,
                    "costs_with_green_hours_target": Z2,
                }
            ),
        }
    )

    logged_costs = LoggedCost.objects.order_by().filter(rendered_on__range=date_range)

    where = ["lh.rendered_on >= %s and lh.rendered_on <= %s"]
    params = date_range[:]

    if project is not None:
        where.append("p.id=%s")
        params.append(project)
        logged_costs = logged_costs.filter(service__project=project)
    if cost_center is not None:
        where.append("p.cost_center_id=%s")
        params.append(cost_center)
        logged_costs = logged_costs.filter(service__project__cost_center=cost_center)

    with connections["default"].cursor() as cursor:
        cursor.execute(LABOR_COSTS_SQL % " and ".join(where), params)

        for project_id, hlc, ght, rendered_by_id, hours in cursor:
            project = projects[project_id]
            project["hours"] += hours

            by_user = project["by_user"][rendered_by_id]
            by_user["hours"] += hours

            if hlc is None:
                project["hours_with_rate_undefined"] += hours
                by_user["hours_with_rate_undefined"] += hours

            else:
                costs = hours * hlc
                costs_with_ght = hours * hlc * 100 / ght

                project["costs"] += costs
                project["costs_with_green_hours_target"] += costs_with_ght

                by_user["costs"] += costs
                by_user["costs_with_green_hours_target"] += costs_with_ght

    with connections["default"].cursor() as cursor:
        cursor.execute(REVENUE_SQL % " and ".join(where), params)

        for project_id, revenue in cursor:
            projects[project_id]["revenue"] += revenue

    for row in (
        logged_costs.filter(third_party_costs__isnull=False)
        .values("service__project")
        .annotate(cost=Sum("third_party_costs"))
    ):
        projects[row["service__project"]]["third_party_costs"] += row["cost"]

    for row in logged_costs.values("service__project").annotate(cost=Sum("cost")):
        projects[row["service__project"]]["revenue"] += row["cost"]

    return projects


def labor_costs_by_cost_center(date_range):
    projects = _labor_costs_by_project_id(date_range)

    sorted_projects = sorted(
        [
            {"project": project, **projects[project.id]}
            for project in Project.objects.filter(
                id__in=projects.keys()
            ).select_related("cost_center", "owned_by")
        ],
        key=lambda row: (row["project"].cost_center_id or 1e100, -row["costs"]),
    )

    cost_centers = []
    for cost_center, cc_projects in groupby(
        sorted_projects, lambda row: row["project"].cost_center
    ):
        cc_projects = list(cc_projects)
        cc_row = {key: sum(row[key] for row in cc_projects) for key in PROJECT_KEYS}
        cc_row.update({"cost_center": cost_center, "projects": cc_projects})
        cost_centers.append(cc_row)

    ret = {key: sum(row[key] for row in cost_centers) for key in PROJECT_KEYS}
    ret["cost_centers"] = cost_centers
    return ret


def labor_costs_by_user(date_range, *, project=None, cost_center=None):
    projects = _labor_costs_by_project_id(
        date_range, project=project, cost_center=cost_center
    )

    users = User.objects.filter(
        id__in=set(
            chain.from_iterable(row["by_user"].keys() for row in projects.values())
        )
    )
    by_user = defaultdict(lambda: dict.fromkeys(USER_KEYS, Z2))
    overall = dict.fromkeys(PROJECT_KEYS, Z2)

    for row in projects.values():
        for key in PROJECT_KEYS:
            overall[key] += row[key]
        for key in USER_KEYS:
            for user in users:
                by_user[user][key] += row["by_user"][user.id][key]

    return {"by_user": [{"user": user, **by_user[user]} for user in users], **overall}


def test():  # pragma: no cover
    import datetime as dt
    from pprint import pprint

    pprint(labor_costs_by_cost_center([dt.date(2019, 1, 1), dt.date.today()]))
    # pprint(labor_costs_by_user([dt.date(2020, 1, 1), dt.date.today()], cost_center=1))



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


default_app_config = "workbench.contacts.Config"


class Config(AppConfig):
    name = "workbench.contacts"
    verbose_name = _("contacts")



# Generated by Django 2.2.3 on 2019-07-19 11:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("expenses", "0002_expensereport_closed_on")]

    operations = [
        migrations.RunSQL("SELECT audit_audit_table('expenses_expensereport')", "")
    ]



import re

from django.shortcuts import render
from django.urls import reverse


FALLBACKS = None


def initialize_fallbacks():
    global FALLBACKS

    if FALLBACKS:
        return

    from workbench.invoices.models import Invoice, RecurringInvoice
    from workbench.offers.models import Offer
    from workbench.projects.models import Project

    FALLBACKS = [
        (
            r"^{}(?P<id>[0-9]+)/".format(Project.urls["list"]),
            lambda kw: ("projects_project", "id", kw["id"]),
        ),
        (
            r"^{}(?P<id>[0-9]+)/".format(Offer.urls["list"]),
            lambda kw: ("offers_offer", "id", kw["id"]),
        ),
        (
            r"^{}(?P<id>[0-9]+)/".format(Invoice.urls["list"]),
            lambda kw: ("invoices_invoice", "id", kw["id"]),
        ),
        (
            r"^{}(?P<id>[0-9]+)/".format(RecurringInvoice.urls["list"]),
            lambda kw: ("invoices_recurringinvoice", "id", kw["id"]),
        ),
    ]


def history_fallback(get_response):
    initialize_fallbacks()

    def middleware(request):
        response = get_response(request)
        if response.status_code == 404:
            for r, params in FALLBACKS:
                match = re.match(r, request.path_info)
                if match:
                    return render(
                        request,
                        "open-modal.html",
                        {"url": reverse("history", args=params(match.groupdict()))},
                        status=404,
                    )

        return response

    return middleware



# Generated by Django 3.1.2 on 2020-10-17 08:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("accounts", "0010_auto_20201016_0957"),
        ("audit", "0004_auto_20201017_1016"),
    ]

    operations = [
        migrations.RunSQL(
            "SELECT audit_audit_table_initial('accounts_team');",
            "DELETE FROM audit_logged_actions WHERE table_name='accounts_team';",
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-04 21:39

import datetime
from decimal import Decimal

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Absence",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("starts_on", models.DateField(verbose_name="starts on")),
                (
                    "days",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="days"
                    ),
                ),
                ("description", models.TextField(verbose_name="description")),
                (
                    "is_vacation",
                    models.BooleanField(default=True, verbose_name="is vacation"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="absences",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "absence",
                "verbose_name_plural": "absences",
                "ordering": ["-starts_on", "-pk"],
            },
        ),
        migrations.CreateModel(
            name="Employment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "date_from",
                    models.DateField(
                        default=datetime.date.today, verbose_name="date from"
                    ),
                ),
                (
                    "date_until",
                    models.DateField(
                        default=datetime.date(9999, 12, 31), verbose_name="date until"
                    ),
                ),
                ("percentage", models.IntegerField(verbose_name="percentage")),
                (
                    "vacation_weeks",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="Vacation weeks for a full year.",
                        max_digits=4,
                        verbose_name="vacation weeks",
                    ),
                ),
                ("notes", models.TextField(blank=True, verbose_name="notes")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="employments",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "employment",
                "verbose_name_plural": "employments",
                "ordering": ["date_from"],
            },
        ),
        migrations.CreateModel(
            name="Year",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("year", models.IntegerField(unique=True, verbose_name="year")),
                (
                    "january",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="january"
                    ),
                ),
                (
                    "february",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="february"
                    ),
                ),
                (
                    "march",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="march"
                    ),
                ),
                (
                    "april",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="april"
                    ),
                ),
                (
                    "may",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="may"
                    ),
                ),
                (
                    "june",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="june"
                    ),
                ),
                (
                    "july",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="july"
                    ),
                ),
                (
                    "august",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="august"
                    ),
                ),
                (
                    "september",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="september"
                    ),
                ),
                (
                    "october",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="october"
                    ),
                ),
                (
                    "november",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="november"
                    ),
                ),
                (
                    "december",
                    models.DecimalField(
                        decimal_places=2, max_digits=4, verbose_name="december"
                    ),
                ),
                (
                    "working_time_per_day",
                    models.DecimalField(
                        decimal_places=1,
                        max_digits=4,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="working time per day",
                    ),
                ),
            ],
            options={
                "verbose_name": "year",
                "verbose_name_plural": "years",
                "ordering": ["-year"],
            },
        ),
        migrations.AlterUniqueTogether(
            name="employment", unique_together={("user", "date_from")}
        ),
    ]



# Generated by Django 3.0.3 on 2020-03-02 20:04

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contacts", "0012_person_date_of_birth"),
    ]

    operations = [
        migrations.CreateModel(
            name="Attribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
                (
                    "is_archived",
                    models.BooleanField(default=False, verbose_name="is archived"),
                ),
            ],
            options={
                "verbose_name": "attribute",
                "verbose_name_plural": "attributes",
                "ordering": ("position", "id"),
            },
        ),
        migrations.CreateModel(
            name="AttributeGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
                (
                    "is_archived",
                    models.BooleanField(default=False, verbose_name="is archived"),
                ),
                (
                    "is_required",
                    models.BooleanField(default=True, verbose_name="is required"),
                ),
            ],
            options={
                "verbose_name": "attribute group",
                "verbose_name_plural": "attribute groups",
                "ordering": ("position", "id"),
            },
        ),
        migrations.CreateModel(
            name="ClosingType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "represents_a_win",
                    models.BooleanField(default=False, verbose_name="represents a win"),
                ),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
            ],
            options={
                "verbose_name": "closing type",
                "verbose_name_plural": "closing types",
                "ordering": ("position", "id"),
            },
        ),
        migrations.CreateModel(
            name="Deal",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "value",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="value",
                    ),
                ),
                (
                    "status",
                    models.PositiveIntegerField(
                        choices=[(10, "Open"), (20, "Accepted"), (30, "Declined")],
                        default=10,
                        verbose_name="status",
                    ),
                ),
                (
                    "probability",
                    models.IntegerField(
                        choices=[(10, "unknown"), (20, "normal"), (30, "high")],
                        default=10,
                        verbose_name="probability",
                    ),
                ),
                (
                    "decision_expected_on",
                    models.DateField(
                        blank=True, null=True, verbose_name="decision expected on"
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "closed_on",
                    models.DateField(blank=True, null=True, verbose_name="closed on"),
                ),
                (
                    "closing_notice",
                    models.TextField(blank=True, verbose_name="closing notice"),
                ),
                ("_fts", models.TextField(blank=True, editable=False)),
            ],
            options={
                "verbose_name": "deal",
                "verbose_name_plural": "deals",
                "ordering": ["-pk"],
            },
        ),
        migrations.CreateModel(
            name="ValueType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "position",
                    models.PositiveIntegerField(default=0, verbose_name="position"),
                ),
                (
                    "is_archived",
                    models.BooleanField(default=False, verbose_name="is archived"),
                ),
            ],
            options={
                "verbose_name": "value type",
                "verbose_name_plural": "value types",
                "ordering": ("position", "id"),
            },
        ),
        migrations.CreateModel(
            name="DealAttribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "attribute",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="deals.Attribute",
                        verbose_name="attribute",
                    ),
                ),
                (
                    "deal",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="deals.Deal",
                        verbose_name="deal",
                    ),
                ),
            ],
            options={
                "verbose_name": "deal attribute",
                "verbose_name_plural": "deal attributes",
            },
        ),
        migrations.AddField(
            model_name="deal",
            name="attributes",
            field=models.ManyToManyField(
                through="deals.DealAttribute",
                to="deals.Attribute",
                verbose_name="attributes",
            ),
        ),
        migrations.AddField(
            model_name="deal",
            name="closing_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="deals.ClosingType",
                verbose_name="closing type",
            ),
        ),
        migrations.AddField(
            model_name="deal",
            name="contact",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="contacts.Person",
                verbose_name="contact",
            ),
        ),
        migrations.AddField(
            model_name="deal",
            name="customer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="contacts.Organization",
                verbose_name="customer",
            ),
        ),
        migrations.AddField(
            model_name="deal",
            name="owned_by",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
                verbose_name="responsible",
            ),
        ),
        migrations.AddField(
            model_name="attribute",
            name="group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="attributes",
                to="deals.AttributeGroup",
                verbose_name="attribute group",
            ),
        ),
        migrations.CreateModel(
            name="Value",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "value",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="value",
                    ),
                ),
                (
                    "deal",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values",
                        to="deals.Deal",
                        verbose_name="deal",
                    ),
                ),
                (
                    "type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="deals.ValueType",
                        verbose_name="type",
                    ),
                ),
            ],
            options={
                "verbose_name": "value",
                "verbose_name_plural": "values",
                "ordering": ["type"],
                "unique_together": {("deal", "type")},
            },
        ),
    ]



from functools import wraps

from django.conf import settings
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.utils.translation import gettext as _


def feature_required(feature):
    def decorator(view):
        @wraps(view)
        def require_feature(request, *args, **kwargs):
            if request.user.features[feature]:
                return view(request, *args, **kwargs)
            messages.warning(request, _("Feature not available"))
            return HttpResponseRedirect("/")

        return require_feature

    return decorator


class FEATURES:
    BOOKKEEPING = "bookkeeping"
    CAMPAIGNS = "campaigns"
    CONTROLLING = "controlling"
    DEALS = "deals"
    FOREIGN_CURRENCIES = "foreign_currencies"
    GLASSFROG = "glassfrog"
    LABOR_COSTS = "labor_costs"
    PLANNING = "planning"
    SKIP_BREAKS = "skip_breaks"
    WORKING_TIME_CORRECTION = "working_time_correction"


bookkeeping_only = feature_required(FEATURES.BOOKKEEPING)
controlling_only = feature_required(FEATURES.CONTROLLING)
deals_only = feature_required(FEATURES.DEALS)
labor_costs_only = feature_required(FEATURES.LABOR_COSTS)


KNOWN_FEATURES = {getattr(FEATURES, attr) for attr in dir(FEATURES) if attr.isupper()}


class UnknownFeature(Exception):
    pass


class UserFeatures:
    def __init__(self, *, email):
        self.email = email

    def __getattr__(self, key):
        try:
            setting = settings.FEATURES[key]
        except KeyError:
            if key not in KNOWN_FEATURES:
                raise UnknownFeature("Unknown feature: %r" % key)

            return False
        if setting is True or setting is False:
            return setting
        return self.email in setting

    __getitem__ = __getattr__



from workbench.tools import admin

from . import models


@admin.register(models.LoggedHours)
class LoggedHoursAdmin(admin.ModelAdmin):
    list_display = (
        "service",
        "created_at",
        "created_by",
        "rendered_on",
        "rendered_by",
        "hours",
        "description",
    )
    raw_id_fields = ("service", "invoice_service")


@admin.register(models.LoggedCost)
class LoggedCostAdmin(admin.ModelAdmin):
    list_display = ("service", "created_by", "rendered_on", "cost", "description")
    raw_id_fields = ("service", "invoice_service", "expense_report")


@admin.register(models.Break)
class BreakAdmin(admin.ModelAdmin):
    list_display = ["user", "starts_at", "ends_at", "description"]
    raw_id_fields = ["user"]



import datetime as dt

from django.test import RequestFactory, TestCase
from django.test.utils import override_settings
from django.urls import reverse

from workbench import factories
from workbench.accounts.features import UnknownFeature
from workbench.accounts.forms import TeamForm, TeamSearchForm
from workbench.accounts.models import User
from workbench.projects.models import Project
from workbench.tools.testing import messages


class AccountsTest(TestCase):
    def test_user(self):
        """User's methods and atttributes work as expected"""
        with self.assertRaises(ValueError):
            User.objects.create_user("", "")

        user = User.objects.create_user("test@example.com", "")
        self.assertFalse(user.has_usable_password())
        self.assertTrue(user.has_perm("stuff"))
        self.assertTrue(user.has_module_perms("stuff"))
        self.assertFalse(user.is_staff)

    def test_admin(self):
        """Admin user's methods and atttributes work as expected"""
        user = User.objects.create_superuser("test@example.com", "")
        self.assertFalse(user.has_usable_password())
        self.assertTrue(user.has_perm("stuff"))
        self.assertTrue(user.has_module_perms("stuff"))
        self.assertTrue(user.is_staff)

        self.client.force_login(user)

        # Contains the signed email value
        response = self.client.get("/admin/accounts/user/{}/change/".format(user.pk))
        self.assertContains(response, '<div class="readonly">{}:'.format(user.email))

        # Creating an user through the admin interface is possible
        wtm = factories.WorkingTimeModelFactory.create()
        data = {
            "email": "test@example.org",
            "language": "en",
            "working_time_model": wtm.id,
            "planning_hours_per_day": 6,
            "employments-TOTAL_FORMS": 0,
            "employments-INITIAL_FORMS": 0,
            "employments-MAX_NUM_FORMS": 1000,
        }
        response = self.client.post("/admin/accounts/user/add/", data)
        self.assertEqual(response.status_code, 302)

    def test_choices(self):
        """User.objects.choices() does what it should"""
        u1 = factories.UserFactory.create(_full_name="M A", is_active=True)
        u2 = factories.UserFactory.create(_full_name="M I", is_active=False)

        self.assertEqual(
            list(User.objects.choices(collapse_inactive=False)),
            [
                ("", "Alle Nutzer*innen"),
                ("Aktiv", [(u1.pk, "M A")]),
                ("Inaktiv", [(u2.pk, "M I")]),
            ],
        )

        self.assertEqual(
            list(User.objects.choices(collapse_inactive=True)),
            [
                ("", "Alle Nutzer*innen"),
                (0, "Inaktive Nutzer*innen"),
                ("Aktiv", [(u1.pk, "M A")]),
            ],
        )

    def test_ordering(self):
        """Ordering of users while falling back to different attributes"""
        self.assertTrue(User(_short_name="a") < User(_full_name="b"))

    @override_settings(FEATURES={"glassfrog": False})
    def test_feature_required(self):
        """The feature_required decorator redirects users and adds a message"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get(reverse("report_hours_by_circle"))
        self.assertRedirects(response, "/")
        self.assertEqual(messages(response), ["Feature not available"])

    @override_settings(
        FEATURES={"yes": True, "no": False, "maybe": {"test@example.com"}}
    )
    def test_user_features(self):
        """Features may either be enabled for all, for some or for no users"""
        user = User(email="test@example.org")
        self.assertTrue(user.features["yes"])
        self.assertFalse(user.features["no"])
        self.assertFalse(user.features["maybe"])
        with self.assertRaises(UnknownFeature):
            user.features["missing"]

        user = User(email="test@example.com")
        self.assertTrue(user.features["yes"])
        self.assertFalse(user.features["no"])
        self.assertTrue(user.features["maybe"])
        with self.assertRaises(UnknownFeature):
            user.features["missing"]

    def test_user_views(self):
        """The user views do not crash"""
        hours = factories.LoggedHoursFactory.create()
        hours = factories.LoggedHoursFactory.create(
            service=factories.ServiceFactory.create(
                project=factories.ProjectFactory.create(type=Project.INTERNAL)
            ),
            rendered_by=hours.rendered_by,
        )
        self.client.force_login(hours.rendered_by)
        user = hours.rendered_by

        response = self.client.get("/users/")
        self.assertContains(response, user.get_full_name())

        response = self.client.get("/users/{}/".format(user.id))
        self.assertContains(response, user.email)

        response = self.client.get("/users/{}/statistics/".format(user.id))
        self.assertContains(response, "Hours per week")

    def test_work_anniversaries(self):
        """The work anniversaries report doesn't crash"""
        user = factories.UserFactory.create()
        user.employments.create(
            percentage=100, vacation_weeks=5, date_from=dt.date(2018, 1, 1)
        )

        self.client.force_login(user)
        response = self.client.get("/report/work-anniversaries/")
        self.assertContains(response, user.get_full_name())

    def test_team_crud(self):
        """CRUD of teams"""

        rf = RequestFactory()
        req = rf.get("/")

        form = TeamSearchForm(request=req)
        stuff = object()
        self.assertEqual(form.filter(stuff), stuff)

        user = factories.UserFactory.create()
        inactive = factories.UserFactory.create(is_active=False)

        form = TeamForm({"name": "A Team", "members": [user.id]}, request=req)
        self.assertEqual(set(form.fields["members"].queryset), {user})

        team = form.save()
        self.assertEqual(set(team.members.all()), {user})

        team.members.add(inactive)
        form = TeamForm(instance=team, request=req)
        self.assertEqual(set(form.fields["members"].queryset), {user, inactive})



# Generated by Django 3.1a1 on 2020-05-25 10:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("offers", "0008_auto_20190916_1110"),
    ]

    operations = [
        migrations.AddField(
            model_name="offer",
            name="valid_until",
            field=models.DateField(blank=True, null=True, verbose_name="valid until"),
        ),
        migrations.RunSQL(
            "UPDATE offers_offer SET valid_until=offered_on + interval '59 days'"
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-08 16:52

from decimal import Decimal

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("invoices", "0005_auto_20190307_0925")]

    operations = [
        migrations.CreateModel(
            name="CreditEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "reference_number",
                    models.CharField(
                        max_length=40, unique=True, verbose_name="reference number"
                    ),
                ),
                ("value_date", models.DateField(verbose_name="value date")),
                (
                    "total",
                    models.DecimalField(
                        decimal_places=2,
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="total",
                    ),
                ),
                (
                    "payment_notice",
                    models.CharField(
                        blank=True, max_length=1000, verbose_name="payment notice"
                    ),
                ),
                ("notes", models.TextField(blank=True, verbose_name="notes")),
                (
                    "invoice",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="invoices.Invoice",
                        verbose_name="invoice",
                    ),
                ),
            ],
            options={
                "verbose_name": "credit entry",
                "verbose_name_plural": "credit entries",
                "ordering": ["-value_date", "-pk"],
            },
        )
    ]



import time
from functools import wraps
from urllib.parse import urlencode

from django import forms
from django.forms.utils import flatatt
from django.urls import reverse
from django.utils.encoding import force_str
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe
from django.utils.text import capfirst
from django.utils.translation import gettext, gettext_lazy as _

from workbench.accounts.models import User


class Textarea(forms.Textarea):
    def __init__(self, attrs=None):
        default_attrs = {"cols": 40, "rows": 5}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(default_attrs)


_AUTOCOMPLETE_TEMPLATE = """
<div class="input-group">
  <input type="text" class="form-control" id="%(id)s_pretty"
    value="%(pretty)s" placeholder="%(placeholder)s"
    autocomplete="off"
    data-autocomplete-url="%(url)s" data-autocomplete-id="%(id)s">
  <div class="input-group-append">
    <button type="button" class="btn btn-primary" %(btn_attrs)s
      data-clear="#%(id)s,#%(id)s_pretty">&times;</button>
  </div>
</div>
%(field)s
"""  # noqa


class Autocomplete(forms.TextInput):
    def __init__(self, model, attrs=None, *, params=None):
        super().__init__(attrs)
        self.model = model
        self.params = ("?" + urlencode(params)) if params else ""

    def render(self, name, value, attrs=None, choices=(), renderer=None):
        if value is None:
            value = ""
        final_attrs = self.build_attrs(attrs, {"type": "hidden", "name": name})
        if value != "":
            # Only add the 'value' attribute if a value is non-empty.
            final_attrs["value"] = force_str(self.format_value(value))

        pretty = ""
        try:
            if value:
                pretty = str(self.model.objects.get(pk=value))
        except (self.model.DoesNotExist, TypeError, ValueError):
            pass

        if final_attrs.get("disabled"):
            return super().render(name, pretty or value, attrs, renderer=renderer)

        opts = self.model._meta
        return mark_safe(
            _AUTOCOMPLETE_TEMPLATE
            % {
                "id": final_attrs["id"],
                "url": reverse("%s_%s_autocomplete" % (opts.app_label, opts.model_name))
                + self.params,
                "field": format_html("<input{} />", flatatt(final_attrs)),
                "pretty": escape(pretty),
                "placeholder": capfirst(opts.verbose_name),
                "btn_attrs": "" if value else "disabled",
            }
        )


class WarningsForm:
    """
    Form class mixin which allows implementing validation warnings
    In contrast to Django's ``ValidationError``, these warnings may
    be ignored by checking a checkbox.
    The warnings support consists of the following methods and properties:
    * ``WarningsForm.add_warning(<warning>)``: Adds a new warning message
    * ``WarningsForm.warnings``: A list of warnings or an empty list if there
      are none.
    * ``WarningsForm.is_valid()``: Overridden ``Form.is_valid()``
      implementation which returns ``False`` for otherwise valid forms with
      warnings, if those warnings have not been explicitly ignored (by checking
      a checkbox or by passing ``ignore_warnings=True`` to ``is_valid()``.
    * An additional form field named ``ignore_warnings`` is available - this
      field should only be displayed if ``WarningsForm.warnings`` is non-emtpy.
    """

    def __init__(self, *args, **kwargs):
        super(WarningsForm, self).__init__(*args, **kwargs)

        self.warnings = {}

    def add_warning(self, warning, *, code):
        """
        Adds a new warning, should be called while cleaning the data
        """
        self.warnings[code] = warning

    def is_valid(self, ignore_warnings=False):
        """
        ``is_valid()`` override which returns ``False`` for forms with warnings
        if these warnings haven't been explicitly ignored
        """
        if not super(WarningsForm, self).is_valid():
            return False

        if self.warnings and not self.should_ignore_warnings():
            return False

        return True

    def should_ignore_warnings(self):
        ignore = set(self.data.get(self.ignore_warnings_id, "").split())
        current = set(self.warnings)
        # print("IGNORE", ignore, "CURRENT", current)
        return current <= ignore

    @property
    def ignore_warnings_value(self):
        return " ".join(sorted(self.warnings))

    ignore_warnings_id = "__ig_{}".format(int(time.time() / 10800))


class DateInput(forms.TextInput):
    input_type = "date"


class Form(WarningsForm, forms.Form):
    required_css_class = "required"
    error_css_class = "is-invalid"

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def apply_simple(self, queryset, *fields):
        data = self.cleaned_data
        for field, value in ((field, data.get(field)) for field in fields):
            if value:
                queryset = queryset.filter(**{field: value})
        return queryset

    def apply_renamed(self, queryset, param, field):
        value = self.cleaned_data.get(param)
        if value:
            return queryset.filter(**{field: value})
        return queryset

    def apply_owned_by(self, queryset, *, attribute="owned_by"):
        value = self.cleaned_data.get(attribute)
        if value == 0:
            return queryset.filter(**{"{}__is_active".format(attribute): False})
        elif value:
            user = self.request.user if value == -1 else value
            return queryset.filter(**{attribute: user})
        return queryset


class ModelForm(WarningsForm, forms.ModelForm):
    user_fields = ()
    default_to_current_user = ()
    required_css_class = "required"
    error_css_class = "is-invalid"

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")

        if self.default_to_current_user:
            instance = kwargs.get("instance")
            if not instance or not instance.pk:
                initial = kwargs.setdefault("initial", {})
                for field in self.default_to_current_user:
                    initial.setdefault(field, self.request.user.pk)

        super().__init__(*args, **kwargs)

        for name, field in self.fields.items():
            if isinstance(field, forms.DateField):
                field.widget = DateInput()

            elif name in self.user_fields:
                field.choices = User.objects.active_choices(
                    include=self.instance
                    and getattr(self.instance, "%s_id" % name, None)
                )

        self.customer_and_contact = all(
            f in self.fields for f in ("customer", "contact")
        )
        if self.customer_and_contact:
            self.fields["customer"].required = False
            self.fields["customer"].help_text = self.fields["customer"].help_text or _(
                "Is automatically filled using the organization's contact."
            )

    def clean(self):
        data = super().clean()

        if self.customer_and_contact:
            if data.get("contact") and not data.get("customer"):
                data["customer"] = data["contact"].organization

            if data.get("customer") and data.get("contact"):
                if data.get("customer") != data.get("contact").organization:
                    raise forms.ValidationError(
                        {
                            "contact": gettext(
                                "The contact %(person)s does not belong to"
                                " %(organization)s."
                            )
                            % {
                                "person": data.get("contact"),
                                "organization": data.get("customer"),
                            }
                        }
                    )

            if not data.get("customer") and "customer" in self.fields:
                raise forms.ValidationError(
                    {"customer": self.fields["customer"].error_messages["required"]},
                    code="required",
                )

            if not data.get("contact"):
                self.add_warning(_("No contact selected."), code="no-contact")

        return data


def add_prefix(prefix):
    def decorator(cls):
        @wraps(cls)
        def fn(*args, **kwargs):
            kwargs.setdefault("prefix", prefix)
            return cls(*args, **kwargs)

        return fn

    return decorator


def querystring(params=None, **kwargs):
    params = params.items() if params else ()
    query = urlencode(
        sorted(
            (key, value)
            for key, value in dict(params, **kwargs).items()
            if key not in {"error"} and value
        )
    )
    return "?%s" % query



import json

from django.contrib import messages
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.decorators import classonlymethod
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import gettext as _, gettext_lazy

import vanilla

from workbench.services.models import ServiceType


class ToolsMixin(object):
    title = None

    @classonlymethod
    def as_view(cls, **initkwargs):
        assert cls.model or initkwargs.get("model"), "model is required for view"
        return super().as_view(**initkwargs)

    @cached_property
    def default_service_types(self):
        return ServiceType.objects.all()

    @cached_property
    def default_service_types_json(self):
        return json.dumps(
            {
                str(type.id): {
                    "effort_type": type.title,
                    "effort_rate": int(type.hourly_rate),
                }
                for type in self.default_service_types
            }
        )

    def get_template_names(self):
        """
        Returns a list of template names to use when rendering the response.
        If `.template_name` is not specified, then defaults to the following
        pattern: "{app_label}/{model_name}{template_name_suffix}.html"
        """
        if self.template_name is not None:
            return [self.template_name]
        return [
            "%s/%s%s.html"
            % (
                self.model._meta.app_label,
                self.model._meta.object_name.lower(),
                self.template_name_suffix,
            ),
            "generic/object%s.html" % self.template_name_suffix,
        ]

    @property
    def meta(self):
        return self.model._meta

    def get_form(self, data=None, files=None, **kwargs):
        kwargs["request"] = self.request
        cls = self.get_form_class()
        return cls(data=data, files=files, **kwargs)

    def get_context_data(self, **kwargs):
        if self.title:
            kwargs.setdefault(
                "title",
                capfirst(
                    self.title
                    % {
                        "object": self.model._meta.verbose_name,
                        "instance": getattr(self, "object", None),
                    }
                ),
            )
        return super().get_context_data(**kwargs)


class ListView(ToolsMixin, vanilla.ListView):
    paginate_by = 100
    search_form_class = None
    show_create_button = True

    @classonlymethod
    def as_view(cls, **initkwargs):
        assert cls.search_form_class or initkwargs.get(
            "search_form_class"
        ), "search_form_class is required for list view"
        return super().as_view(**initkwargs)

    def get(self, request, *args, **kwargs):
        self.search_form = self.search_form_class(request.GET, request=request)
        if not self.search_form.is_valid():
            messages.warning(request, _("Search form was invalid."))
            return HttpResponseRedirect("?error=1")

        if (
            set(request.GET)
            - set(self.search_form.fields)
            # Also see workbench.js
            - {"disposition", "error", "export", "page"}
        ):
            messages.warning(request, _("Invalid parameters to form."))
            return HttpResponseRedirect("?error=1")

        if hasattr(self.search_form, "response"):
            response = self.search_form.response(request, self.get_queryset())
            if response:
                return response
        return super().get(request, *args, **kwargs)

    def get_queryset(self):
        return self.search_form.filter(super().get_queryset())


class DetailView(ToolsMixin, vanilla.DetailView):
    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        url = self.model.get_redirect_url(self.object, request)
        if url:
            return redirect(url)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return self.render_to_response(context)


class CreateView(ToolsMixin, vanilla.CreateView):
    title = gettext_lazy("Create %(object)s")

    def dispatch(self, request, *args, **kwargs):
        if not self.model.allow_create(request):
            return (
                render(request, "modal.html") if request.is_ajax() else redirect("../")
            )
        url = self.model.get_redirect_url(None, request)
        if url:
            return redirect(url)
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        self.object = form.save()
        messages.success(
            self.request,
            capfirst(
                _("%(class)s '%(object)s' has been created successfully.")
                % {"class": self.object._meta.verbose_name, "object": self.object}
            ),
        )

        if self.request.is_ajax():
            return HttpResponse("Thanks", status=201)  # Created
        return redirect(self.get_success_url())


class CreateAndUpdateView(CreateView):
    def get_success_url(self):
        return self.object.urls["update"]


class CreateRelatedView(CreateView):
    related_model = None

    def get_form(self, *args, **kwargs):
        instance = get_object_or_404(self.related_model, pk=self.kwargs.pop("pk"))
        attribute = self.related_model.__name__.lower()
        setattr(self, attribute, instance)
        kwargs[attribute] = instance
        return super().get_form(*args, **kwargs)


class UpdateView(ToolsMixin, vanilla.UpdateView):
    title = gettext_lazy("Update %(object)s")

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.object.allow_update(self.object, request):
            return (
                render(request, "modal.html")
                if request.is_ajax()
                else redirect(self.object)
            )
        url = self.model.get_redirect_url(self.object, request)
        if url:
            return redirect(url)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        form = self.get_form(instance=self.object)
        context = self.get_context_data(form=form)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        form = self.get_form(
            data=request.POST, files=request.FILES, instance=self.object
        )
        if form.is_valid():
            return self.form_valid(form)
        return self.form_invalid(form)

    def form_valid(self, form):
        self.object = form.save()
        messages.success(
            self.request,
            capfirst(
                _("%(class)s '%(object)s' has been updated successfully.")
                % {"class": self.object._meta.verbose_name, "object": self.object}
            ),
        )
        if self.request.is_ajax():
            return HttpResponse("Thanks", status=202)  # Accepted
        return redirect(self.get_success_url())


class DeleteView(ToolsMixin, vanilla.DeleteView):
    delete_form_class = None
    title = gettext_lazy("Delete %(object)s")

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        allow_delete = self.object.allow_delete(self.object, request)
        if allow_delete is True:
            pass  # Fine!
        elif allow_delete is False:
            return (
                render(request, "modal.html")
                if request.is_ajax()
                else redirect(self.object)
            )
        else:
            assert (
                self.delete_form_class
            ), "delete_form_class must be set if allow_delete returns None"

        url = self.model.get_redirect_url(self.object, request)
        if url:
            return redirect(url)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        if self.delete_form_class:
            form = self.delete_form_class(
                request.POST, instance=self.object, request=request
            )
            if form.is_valid():
                form.delete()
            else:
                context = self.get_context_data()
                context["form"] = form
                return self.render_to_response(context)
        else:
            self.object.delete()

        messages.success(
            self.request,
            capfirst(
                _("%(class)s '%(object)s' has been deleted successfully.")
                % {"class": self.model._meta.verbose_name, "object": self.object},
            ),
        )
        if request.is_ajax():
            return HttpResponse("Thanks", status=204)  # No content
        return redirect(self.get_success_url())

    def get_context_data(self, **kwargs):
        if self.delete_form_class and self.request.method == "GET":
            kwargs["form"] = self.delete_form_class(
                instance=self.object, request=self.request
            )
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return self.model.urls["list"]


class AutocompleteView(ToolsMixin, vanilla.ListView):
    filter = None
    label_from_instance = str

    def get(self, request, *args, **kwargs):
        q = request.GET.get("q")
        queryset = self.get_queryset().search(q)
        queryset = (
            self.filter(queryset=queryset, request=request) if self.filter else queryset
        )
        return JsonResponse(
            {
                "results": [
                    {"label": self.label_from_instance(instance), "value": instance.pk}
                    for instance in queryset[:50]
                ]
                if q
                else []
            }
        )



# Generated by Django 3.0 on 2019-12-07 12:04

from django.db import migrations


def forwards(apps, schema_editor):
    wtm = apps.get_model("awt", "WorkingTimeModel").objects.create(name="Zürich")
    apps.get_model("awt", "Year").objects.update(working_time_model=wtm.id)
    apps.get_model("accounts", "User").objects.update(working_time_model=wtm.id)


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0005_auto_20191207_1304"),
        ("accounts", "0005_user_working_time_model"),
    ]

    operations = [migrations.RunPython(forwards)]



from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from workbench.credit_control import parsers
from workbench.invoices.models import Invoice
from workbench.tools.models import Model, MoneyField, SearchQuerySet
from workbench.tools.urls import model_urls


# @model_urls
class Ledger(Model):
    PARSER_CHOICES = [
        ("zkb-csv", _("ZKB CSV")),
        ("postfinance-csv", _("PostFinance CSV")),
    ]

    name = models.CharField(_("name"), max_length=100)
    parser = models.CharField(_("parser"), max_length=20, choices=PARSER_CHOICES)

    class Meta:
        ordering = ["name"]
        verbose_name = _("ledger")
        verbose_name_plural = _("ledgers")

    def __str__(self):
        return self.name

    @property
    def parse_fn(self):
        return {
            "zkb-csv": parsers.parse_zkb_csv,
            "postfinance-csv": parsers.parse_postfinance_csv,
        }[self.parser]


class CreditEntryQuerySet(SearchQuerySet):
    def pending(self):
        return self.filter(Q(invoice__isnull=True) & Q(notes=""))

    def processed(self):
        return self.filter(~Q(invoice__isnull=True) | ~Q(notes=""))


@model_urls
class CreditEntry(Model):
    ledger = models.ForeignKey(
        Ledger,
        on_delete=models.PROTECT,
        related_name="transactions",
        verbose_name=_("ledger"),
    )
    reference_number = models.CharField(
        _("reference number"), max_length=40, unique=True
    )
    value_date = models.DateField(_("value date"))
    total = MoneyField(_("total"))
    payment_notice = models.CharField(_("payment notice"), max_length=1000, blank=True)

    invoice = models.OneToOneField(
        Invoice,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("invoice"),
    )
    notes = models.TextField(_("notes"), blank=True)
    _fts = models.TextField(editable=False, blank=True)

    objects = CreditEntryQuerySet.as_manager()

    class Meta:
        ordering = ["-value_date", "-pk"]
        verbose_name = _("credit entry")
        verbose_name_plural = _("credit entries")

    def __str__(self):
        return self.reference_number

    def save(self, *args, **kwargs):
        self._fts = " ".join(str(part) for part in [self.invoice or "", self.total])
        super().save(*args, **kwargs)

    save.alters_data = True



from django.urls import path

from workbench import generic
from workbench.logbook.views import create
from workbench.planning.forms import (
    PlannedWorkForm,
    PlannedWorkSearchForm,
    PlanningRequestForm,
    PlanningRequestSearchForm,
)
from workbench.planning.models import PlannedWork, PlanningRequest
from workbench.projects.models import Project


urlpatterns = [
    # Requests
    path(
        "requests/",
        generic.ListView.as_view(
            model=PlanningRequest, search_form_class=PlanningRequestSearchForm
        ),
        name="planning_planningrequest_list",
    ),
    path(
        "requests/<int:pk>/",
        generic.DetailView.as_view(model=PlanningRequest),
        name="planning_planningrequest_detail",
    ),
    path(
        "requests/create/",
        create,
        {"viewname": "createrequest"},
        name="planning_planningrequest_create",
    ),
    path(
        "requests/create/<int:pk>/",
        generic.CreateRelatedView.as_view(
            model=PlanningRequest, form_class=PlanningRequestForm, related_model=Project
        ),
        name="projects_project_createrequest",
    ),
    path(
        "requests/<int:pk>/update/",
        generic.UpdateView.as_view(
            model=PlanningRequest,
            form_class=PlanningRequestForm,
        ),
        name="planning_planningrequest_update",
    ),
    path(
        "requests/<int:pk>/delete/",
        generic.DeleteView.as_view(
            model=PlanningRequest, template_name="modal_confirm_delete.html"
        ),
        name="planning_planningrequest_delete",
    ),
    # Planned work
    path(
        "work/",
        generic.ListView.as_view(
            model=PlannedWork, search_form_class=PlannedWorkSearchForm
        ),
        name="planning_plannedwork_list",
    ),
    path(
        "work/<int:pk>/",
        generic.DetailView.as_view(model=PlannedWork),
        name="planning_plannedwork_detail",
    ),
    path(
        "work/create/",
        create,
        {"viewname": "creatework"},
        name="planning_plannedwork_create",
    ),
    path(
        "work/create/<int:pk>/",
        generic.CreateRelatedView.as_view(
            model=PlannedWork, form_class=PlannedWorkForm, related_model=Project
        ),
        name="projects_project_creatework",
    ),
    path(
        "work/<int:pk>/update/",
        generic.UpdateView.as_view(model=PlannedWork, form_class=PlannedWorkForm),
        name="planning_plannedwork_update",
    ),
    path(
        "work/<int:pk>/delete/",
        generic.DeleteView.as_view(
            model=PlannedWork, template_name="modal_confirm_delete.html"
        ),
        name="planning_plannedwork_delete",
    ),
]



from django.shortcuts import get_object_or_404, redirect
from django.urls import include, re_path
from django.utils.translation import gettext_lazy as _

from workbench import generic
from workbench.accounts.features import controlling_only
from workbench.invoices.forms import CreateProjectInvoiceForm
from workbench.invoices.models import Invoice
from workbench.logbook.forms import LoggedCostForm, LoggedHoursForm
from workbench.logbook.models import LoggedCost, LoggedHours
from workbench.offers.forms import OfferForm
from workbench.offers.models import Offer
from workbench.offers.views import ProjectOfferPDFView
from workbench.planning.views import ProjectPlanningView
from workbench.projects.forms import (
    CampaignDeleteForm,
    CampaignForm,
    CampaignSearchForm,
    ProjectForm,
    ProjectSearchForm,
    ReassignLogbookForm,
    ServiceForm,
    ServiceMoveForm,
)
from workbench.projects.models import Campaign, Project, Service
from workbench.projects.views import (
    OffersRenumberView,
    assign_service_type,
    projects,
    select,
    services,
    set_order,
)


def autocomplete_filter(*, request, queryset):
    return (
        queryset.filter(closed_on__isnull=True)
        if request.GET.get("only_open")
        else queryset
    )


urlpatterns = [
    re_path(r"^offers/", include("workbench.offers.urls")),
    # Campaigns
    re_path(
        r"^campaigns/$",
        generic.ListView.as_view(model=Campaign, search_form_class=CampaignSearchForm),
        name="projects_campaign_list",
    ),
    re_path(
        r"^campaigns/autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Campaign,
            queryset=Campaign.objects.select_related("owned_by"),
            filter=autocomplete_filter,
        ),
        name="projects_campaign_autocomplete",
    ),
    re_path(
        r"^campaigns/(?P<pk>\d+)/$",
        generic.DetailView.as_view(
            model=Campaign,
            queryset=Campaign.objects.select_related("owned_by", "customer"),
        ),
        name="projects_campaign_detail",
    ),
    re_path(
        r"^campaigns/create/$",
        generic.CreateView.as_view(form_class=CampaignForm, model=Campaign),
        name="projects_campaign_create",
    ),
    re_path(
        r"^campaigns/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(form_class=CampaignForm, model=Campaign),
        name="projects_campaign_update",
    ),
    re_path(
        r"^campaigns/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=Campaign, delete_form_class=CampaignDeleteForm
        ),
        name="projects_campaign_delete",
    ),
    # Projects
    re_path(
        r"^$",
        generic.ListView.as_view(model=Project, search_form_class=ProjectSearchForm),
        name="projects_project_list",
    ),
    re_path(
        r"^autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Project,
            queryset=Project.objects.select_related("owned_by"),
            filter=autocomplete_filter,
        ),
        name="projects_project_autocomplete",
    ),
    re_path(r"^select/$", select, name="projects_project_select"),
    re_path(
        r"^(?P<pk>\d+)/$",
        generic.DetailView.as_view(
            model=Project,
            queryset=Project.objects.select_related(
                "owned_by", "customer", "contact__organization"
            ),
        ),
        name="projects_project_detail",
    ),
    re_path(
        r"^(?P<pk>\d+)/statistics/$",
        generic.DetailView.as_view(model=Project, template_name_suffix="_statistics"),
        name="projects_project_statistics",
    ),
    re_path(
        r"^(?P<pk>\d+)/planning/$",
        ProjectPlanningView.as_view(),
        name="projects_project_planning",
    ),
    re_path(
        r"^(?P<pk>\d+)/offers-pdf/$",
        controlling_only(ProjectOfferPDFView.as_view()),
        name="projects_project_offers_pdf",
    ),
    re_path(
        r"^(?P<pk>\d+)/renumber-offers/$",
        controlling_only(
            OffersRenumberView.as_view(
                model=Project, template_name_suffix="_renumber_offers"
            )
        ),
        name="projects_project_renumber_offers",
    ),
    re_path(
        r"^create/$",
        generic.CreateView.as_view(form_class=ProjectForm, model=Project),
        name="projects_project_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(form_class=ProjectForm, model=Project),
        name="projects_project_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(model=Project),
        name="projects_project_delete",
    ),
    re_path(
        r"^(?P<pk>\d+)/createoffer/$",
        generic.CreateRelatedView.as_view(
            model=Offer, form_class=OfferForm, related_model=Project
        ),
        name="projects_project_createoffer",
    ),
    re_path(
        r"^(?P<pk>\d+)/createinvoice/$",
        generic.CreateRelatedView.as_view(
            model=Invoice, form_class=CreateProjectInvoiceForm, related_model=Project
        ),
        name="projects_project_createinvoice",
    ),
    # HOURS
    re_path(
        r"^(?P<pk>\d+)/createhours/$",
        generic.CreateRelatedView.as_view(
            model=LoggedHours, form_class=LoggedHoursForm, related_model=Project
        ),
        name="projects_project_createhours",
    ),
    # COSTS
    re_path(
        r"^(?P<pk>\d+)/createcost/$",
        generic.CreateRelatedView.as_view(
            model=LoggedCost, form_class=LoggedCostForm, related_model=Project
        ),
        name="projects_project_createcost",
    ),
    # Services
    re_path(
        r"^(?P<pk>\d+)/createservice/$",
        generic.CreateRelatedView.as_view(
            model=Service, form_class=ServiceForm, related_model=Project
        ),
        name="projects_project_createservice",
    ),
    re_path(
        r"^service/(?P<pk>\d+)/$",
        lambda request, pk: redirect(get_object_or_404(Service, pk=pk)),
        name="projects_service_detail",
    ),
    re_path(
        r"^service/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(model=Service, form_class=ServiceForm),
        name="projects_service_update",
    ),
    re_path(
        r"^service/(?P<pk>\d+)/reassign-logbook/$",
        generic.UpdateView.as_view(
            model=Service,
            form_class=ReassignLogbookForm,
            template_name="modalform.html",
            title=_("Reassign logbook entries of %(instance)s"),
        ),
        name="projects_service_reassign_logbook",
    ),
    re_path(
        r"^service/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(
            model=Service, template_name="modal_confirm_delete.html"
        ),
        name="projects_service_delete",
    ),
    re_path(
        r"^service/(?P<pk>\d+)/move/$",
        generic.UpdateView.as_view(
            model=Service, form_class=ServiceMoveForm, template_name="modalform.html"
        ),
        name="projects_service_move",
    ),
    re_path(
        r"^service/(?P<pk>\d+)/assign-service-type/$",
        assign_service_type,
        name="projects_service_assign_service_type",
    ),
    re_path(r"^service/set-order/$", set_order, name="projects_service_set_order"),
    re_path(
        r"^(?P<pk>[0-9]+)/services/$",
        services,
        name="projects_project_services",
    ),
    re_path(r"^projects/$", projects, name="projects_project_projects"),
    re_path(
        r"^service/autocomplete/$",
        generic.AutocompleteView.as_view(
            model=Service,
            queryset=Service.objects.filter(
                project__closed_on__isnull=True
            ).select_related("project__owned_by"),
            label_from_instance=lambda service: "{} ({})".format(
                service, service.project
            ),
        ),
        name="projects_service_autocomplete",
    ),
]



from django.urls import re_path

from workbench import generic
from workbench.expenses.forms import ExpenseReportForm, ExpenseReportSearchForm
from workbench.expenses.models import ExpenseReport
from workbench.expenses.views import ExpenseReportPDFView, convert


urlpatterns = [
    re_path(
        r"^$",
        generic.ListView.as_view(
            model=ExpenseReport, search_form_class=ExpenseReportSearchForm
        ),
        name="expenses_expensereport_list",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=ExpenseReport),
        name="expenses_expensereport_detail",
    ),
    re_path(
        r"^(?P<pk>\d+)/pdf/$",
        ExpenseReportPDFView.as_view(),
        name="expenses_expensereport_pdf",
    ),
    re_path(
        r"^create/$",
        generic.CreateView.as_view(model=ExpenseReport, form_class=ExpenseReportForm),
        name="expenses_expensereport_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(model=ExpenseReport, form_class=ExpenseReportForm),
        name="expenses_expensereport_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(model=ExpenseReport),
        name="expenses_expensereport_delete",
    ),
    re_path(r"^convert/$", convert, name="expenses_convert"),
]



# Generated by Django 3.2 on 2020-06-04 12:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0021_auto_20200324_0829"),
    ]

    operations = [
        migrations.AddField(
            model_name="recurringinvoice",
            name="create_project",
            field=models.BooleanField(
                default=False,
                help_text="Invoices are created without projects by default.",
                verbose_name="Create project?",
            ),
        ),
    ]



import datetime as dt
from itertools import groupby

from django import forms
from django.db.models import Q
from django.shortcuts import render
from django.utils.html import format_html, format_html_join
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import Team, User
from workbench.accounts.reporting import average_employment_duration, work_anniversaries
from workbench.invoices.models import Invoice
from workbench.invoices.utils import next_valid_day
from workbench.logbook.models import LoggedCost
from workbench.logbook.reporting import logbook_stats
from workbench.projects.models import Project
from workbench.projects.reporting import overdrawn_projects
from workbench.reporting import (
    green_hours,
    key_data,
    labor_costs,
    project_budget_statistics,
)
from workbench.reporting.utils import date_ranges
from workbench.tools.formats import Z0, Z2, local_date_format
from workbench.tools.forms import DateInput, Form
from workbench.tools.validation import filter_form, in_days, monday
from workbench.tools.xlsx import WorkbenchXLSXDocument


def overdrawn_projects_view(request):
    return render(
        request,
        "reporting/overdrawn_projects.html",
        {"overdrawn_projects": overdrawn_projects()},
    )


class OpenItemsForm(Form):
    cutoff_date = forms.DateField(label=capfirst(_("cutoff date")), widget=DateInput())

    def __init__(self, data, *args, **kwargs):
        data = data.copy()
        data.setdefault("cutoff_date", dt.date.today().isoformat())
        super().__init__(data, *args, **kwargs)

    def open_items_list(self):
        open_items = (
            Invoice.objects.invoiced()
            .filter(
                Q(invoiced_on__lte=self.cleaned_data["cutoff_date"]),
                Q(closed_on__gt=self.cleaned_data["cutoff_date"])
                | Q(closed_on__isnull=True),
            )
            .order_by("invoiced_on", "pk")
            .select_related("owned_by", "customer", "project")
        )

        return {
            "list": open_items,
            "total_excl_tax": sum((i.total_excl_tax for i in open_items), Z2),
            "total": sum((i.total for i in open_items), Z2),
        }


@filter_form(OpenItemsForm)
def open_items_list(request, form):
    if request.GET.get("export") == "xlsx":
        xlsx = WorkbenchXLSXDocument()
        xlsx.table_from_queryset(
            form.open_items_list()["list"].select_related(
                "customer", "contact__organization", "owned_by", "project__owned_by"
            )
        )
        return xlsx.to_response(
            "open-items-list-{}.xlsx".format(
                form.cleaned_data["cutoff_date"].isoformat()
            )
        )

    return render(
        request,
        "reporting/open_items_list.html",
        {"form": form, "open_items_list": form.open_items_list()},
    )


def key_data_details(fn):
    def view(request, year, month):
        year = int(year)
        month = int(month)
        date_range = [dt.date(year, month, 1)]
        date_range.append(next_valid_day(year, month + 1, 1) - dt.timedelta(days=1))
        return fn(request, date_range)

    return view


@key_data_details
def key_data_gross_profit(request, date_range):
    return render(
        request,
        "reporting/key_data_gross_profit.html",
        {
            "date_range": date_range,
            "invoices": Invoice.objects.invoiced()
            .filter(invoiced_on__range=date_range)
            .order_by("invoiced_on", "id")
            .select_related("project", "owned_by"),
        },
    )


@key_data_details
def key_data_third_party_costs(request, date_range):
    return render(
        request,
        "reporting/key_data_third_party_costs.html",
        {
            "date_range": date_range,
            "third_party_costs": LoggedCost.objects.filter(
                rendered_on__range=date_range,
                third_party_costs__isnull=False,
                invoice_service__isnull=True,
            )
            .order_by("rendered_on", "id")
            .select_related("service"),
            "invoices": Invoice.objects.invoiced()
            .filter(~Q(type=Invoice.DOWN_PAYMENT))
            .filter(Q(invoiced_on__range=date_range), ~Q(third_party_costs=Z2))
            .order_by("invoiced_on", "id")
            .select_related("project", "owned_by"),
        },
    )


def key_data_view(request):
    today = dt.date.today()
    this_months_end = next_valid_day(today.year, today.month + 1, 1) - dt.timedelta(
        days=1
    )
    date_range = [dt.date(this_months_end.year - 3, 1, 1), this_months_end]

    gross_margin_by_month = key_data.gross_margin_by_month(date_range)
    gross_margin_months = {
        row["month"]: row["gross_margin"] for row in gross_margin_by_month
    }

    gross_margin_by_years = {}
    for month in gross_margin_by_month:
        try:
            year = gross_margin_by_years[month["date"].year]
        except KeyError:
            gross_margin_by_years[month["date"].year] = year = {
                "year": month["date"].year,
                "gross_profit": Z2,
                "third_party_costs": Z2,
                "accruals": Z2,
                "gross_margin": Z2,
                "fte": [],
                "margin_per_fte": [],
                "months": [],
            }

        year["months"].append(month)
        year["gross_profit"] += month["gross_profit"]
        year["third_party_costs"] += month["third_party_costs"]
        year["accruals"] += month["accruals"]["delta"]
        year["gross_margin"] += month["gross_margin"]
        year["fte"].append(month["fte"])

    for year in gross_margin_by_years.values():
        year["fte"] = sum(year["fte"]) / len(year["fte"])
        year["margin_per_fte"] = (
            year["gross_margin"] / year["fte"] if year["fte"] else None
        )

    gh = [
        row
        for row in green_hours.green_hours_by_month()
        if date_range[0] <= row["month"] <= date_range[1]
    ]

    def yearly_headline(gh):
        zero = {"green": Z2, "red": Z2, "maintenance": Z2, "internal": Z2, "total": Z2}

        for key, months in groupby(gh, key=lambda row: row["month"].year):
            this = zero.copy()
            months = list(months)
            for month in months:
                this["green"] += month["green"]
                this["red"] += month["red"]
                this["maintenance"] += month["maintenance"]
                this["internal"] += month["internal"]
                this["total"] += month["total"]
            this["percentage"] = (
                100 * (this["green"] + this["maintenance"]) / this["total"]
            ).quantize(Z0)
            yield key, this, months

    return render(
        request,
        "reporting/key_data.html",
        {
            "date_range": date_range,
            "gross_margin_by_years": [
                row[1] for row in sorted(gross_margin_by_years.items())
            ],
            "gross_margin_by_month": gross_margin_by_month,
            "invoiced_corrected": [
                (year, [gross_margin_months.get((year, i), Z2) for i in range(1, 13)])
                for year in range(date_range[0].year, date_range[1].year + 1)
            ],
            "green_hours": yearly_headline(gh),
            "hours_distribution": {
                "labels": [local_date_format(row["month"], fmt="F Y") for row in gh],
                "datasets": [
                    {
                        "label": label,
                        "data": [100 * row[attribute] / row["total"] for row in gh],
                    }
                    for label, attribute in [
                        (_("profitable"), "green"),
                        (_("overdrawn"), "red"),
                        (_("maintenance"), "maintenance"),
                        (_("internal"), "internal"),
                    ]
                ],
            },
            "service_hours_in_open_orders": key_data.service_hours_in_open_orders(),
            "logged_hours_in_open_orders": key_data.logged_hours_in_open_orders(),
            "sent_invoices_total": key_data.sent_invoices_total(),
            "open_offers_total": key_data.open_offers_total(),
            "average_employment_duration": average_employment_duration(),
        },
    )


class ProjectBudgetStatisticsForm(Form):
    owned_by = forms.TypedChoiceField(label="", coerce=int, required=False)
    cutoff_date = forms.DateField(widget=DateInput, label="")
    closed_during_the_last_year = forms.BooleanField(
        label=_("Closed during the last year"), required=False
    )
    internal = forms.BooleanField(label=_("Internal"), required=False)

    def __init__(self, data, *args, **kwargs):
        data = data.copy()
        today = dt.date.today()
        data.setdefault("cutoff_date", today.isoformat())
        kwargs.setdefault("initial", {}).setdefault("cutoff_date", today)
        super().__init__(data, *args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def queryset(self):
        data = self.cleaned_data
        if data.get("closed_during_the_last_year"):
            queryset = Project.objects.closed().filter(closed_on__gte=in_days(-366))
        else:
            queryset = Project.objects.open(on=self.cleaned_data.get("cutoff_date"))
        if data.get("internal"):
            queryset = queryset.filter(type=Project.INTERNAL)
        else:
            queryset = queryset.exclude(type=Project.INTERNAL)
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related("owned_by")


@filter_form(ProjectBudgetStatisticsForm)
def project_budget_statistics_view(request, form):
    statistics = project_budget_statistics.project_budget_statistics(
        form.queryset(), cutoff_date=form.cleaned_data.get("cutoff_date")
    )

    if form.cleaned_data.get("closed_during_the_last_year"):
        statistics["statistics"] = sorted(
            statistics["statistics"], key=lambda s: s["project"].closed_on, reverse=True
        )

    if request.GET.get("export") == "xlsx" and statistics["statistics"]:
        xlsx = WorkbenchXLSXDocument()
        xlsx.project_budget_statistics(statistics)
        return xlsx.to_response("project-budget-statistics.xlsx")

    return render(
        request,
        "reporting/project_budget_statistics.html",
        {"form": form, "statistics": statistics},
    )


class DateRangeFilterForm(Form):
    date_from = forms.DateField(
        label=_("Date from"), required=False, widget=DateInput()
    )
    date_until = forms.DateField(
        label=_("Date until"), required=False, widget=DateInput()
    )

    def __init__(self, data, *args, **kwargs):
        data = data.copy()
        data.setdefault("date_from", monday().isoformat())
        data.setdefault("date_until", (monday() + dt.timedelta(days=6)).isoformat())
        super().__init__(data, *args, **kwargs)

        self.fields["date_from"].help_text = format_html(
            "{}: {}",
            _("Set predefined period"),
            format_html_join(
                ", ", '<a href="#" data-set-period="{}:{}">{}</a>', date_ranges()
            ),
        )


class DateRangeAndTeamFilterForm(DateRangeFilterForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["team"] = forms.TypedChoiceField(
            choices=[("", _("Everyone"))]
            + [
                [
                    capfirst(Team._meta.verbose_name_plural),
                    [(team.id, str(team)) for team in Team.objects.all()],
                ]
            ]
            + [
                [
                    capfirst(User._meta.verbose_name_plural),
                    [
                        (-user.id, str(user))
                        for user in User.objects.filter(is_active=True)
                    ],
                ]
            ],
            label=capfirst(Team._meta.verbose_name),
            required=False,
            coerce=int,
        )

    def users(self):
        data = self.cleaned_data
        queryset = User.objects.all()
        if data.get("team") and data.get("team") > 0:
            queryset = queryset.filter(teams=data.get("team"))
        elif data.get("team") and data.get("team") < 0:
            queryset = queryset.filter(id=-data.get("team"))
        return queryset


@filter_form(DateRangeAndTeamFilterForm)
def hours_filter_view(request, form, *, template_name, stats_fn):
    return render(
        request,
        template_name,
        {
            "form": form,
            "stats": stats_fn(
                [form.cleaned_data["date_from"], form.cleaned_data["date_until"]],
                users=form.users(),
            ),
        },
    )


@filter_form(DateRangeFilterForm)
def labor_costs_view(request, form):
    date_range = [form.cleaned_data["date_from"], form.cleaned_data["date_until"]]

    if project := request.GET.get("project"):
        return render(
            request,
            "reporting/labor_costs_by_user.html",
            {
                "stats": labor_costs.labor_costs_by_user(date_range, project=project),
            },
        )
    elif cost_center := request.GET.get("cost_center"):
        return render(
            request,
            "reporting/labor_costs_by_user.html",
            {
                "stats": labor_costs.labor_costs_by_user(
                    date_range, cost_center=cost_center
                ),
            },
        )
    elif request.GET.get("users"):
        return render(
            request,
            "reporting/labor_costs_by_user.html",
            {"stats": labor_costs.labor_costs_by_user(date_range)},
        )

    return render(
        request,
        "reporting/labor_costs.html",
        {
            "stats": labor_costs.labor_costs_by_cost_center(date_range),
            "date_range": date_range,
            "form": form,
        },
    )


@filter_form(DateRangeFilterForm)
def logging(request, form):
    date_range = [form.cleaned_data["date_from"], form.cleaned_data["date_until"]]

    return render(
        request,
        "reporting/logging.html",
        {"form": form, "logbook_stats": logbook_stats(date_range)},
    )


def work_anniversaries_view(request):
    return render(
        request,
        "reporting/work_anniversaries.html",
        {"work_anniversaries": work_anniversaries()},
    )



# Generated by Django 2.2.5 on 2019-10-02 08:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("invoices", "0018_auto_20190916_1110")]

    operations = [
        migrations.RunSQL(
            """
UPDATE invoices_invoice i
SET
third_party_costs=COALESCE(
    (
        SELECT SUM(third_party_costs)
        FROM invoices_service s
        WHERE
            s.invoice_id=i.id
            AND s.third_party_costs IS NOT NULL
    ),
    0
)
WHERE i.type='services';
                """
        )
    ]
